"""Manage the system hostname using the 'hostname' utility."""

from __future__ import annotations

from collections import deque
from typing import Self

from pytest_mh import MultihostHost, MultihostReentrantUtility
from pytest_mh.conn import ProcessLogLevel

__all__ = ["HostnameUtils"]


class HostnameUtils(MultihostReentrantUtility[MultihostHost]):
    """
    Provides a Python wrapper for the 'hostname' command.

    This is a reentrant utility. When used as a context manager, it will
    automatically save the original hostname on entry and restore it on exit.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        Initialize the HostnameUtils.

        :param host: Remote host instance.
        :type host: MultihostHost object.
        """
        super().__init__(host)
        self.__states: deque[str] = deque()

    def __enter__(self) -> Self:
        """
        Saves current hostname.

        :return: Self.
        :rtype: Self
        """
        self.logger.info("hostname: backing up current name")
        self.__states.append(self.name)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Reverts hostname changes."""
        _name = self.__states.pop()
        self.logger.info(f"hostname: restoring name to '{_name}'")
        self.name = _name

    @property
    def name(self) -> str:
        """
        Gets hostname.

        :return: Current hostname.
        :rtype: str
        """
        self.logger.info("hostname: getting name")
        result = self.host.conn.exec(["hostname"], log_level=ProcessLogLevel.Error)
        return result.stdout.strip()

    @name.setter
    def name(self, name: str) -> None:
        """
        Sets hostname.

        :param name: The new hostname to set.
        :type name: str
        """
        self.logger.info(f"hostname: setting name to '{name}'")
        self.host.conn.exec(["hostname", name], log_level=ProcessLogLevel.Error)

    @property
    def shortname(self) -> str:
        """
        Gets shortname.

        :return: Short hostname.
        :rtype: str
        """
        self.logger.info("hostname: getting short name")
        result = self.host.conn.exec(["hostname", "-s"], log_level=ProcessLogLevel.Error)
        return result.stdout.strip()

    @property
    def fqdn(self) -> str:
        """
        Gets FQDN.

        :return: FQDN.
        :rtype: str
        """
        self.logger.info("hostname: getting FQDN")
        result = self.host.conn.exec(["hostname", "-f"], log_level=ProcessLogLevel.Error)
        return result.stdout.strip()



from __future__ import annotations

import re
from typing import Literal

import pytest

from .. import MultihostArtifactsType, MultihostHost, MultihostUtility
from ..conn import ProcessLogLevel
from .fs import LinuxFileSystem

__all__ = ["Coredumpd"]


class Coredumpd(MultihostUtility[MultihostHost]):
    """
    Coredumpd utilities.

    Collects generated core files from /var/lib/systemd/coredump and if there
    are any, optionally fail the test.

    .. note::

        In order for this functionality to work correctly, systemd-coredumpd
        must be configured to store the core files in a directory (default), not
        in a journal.

        See ``man coredump.conf`` and its ``Storage`` option for more
        information.
    """

    def __init__(
        self,
        host: MultihostHost,
        fs: LinuxFileSystem,
        *,
        mode: Literal["fail", "warn", "ignore"] = "ignore",
        filter: str | None = None,
        path: str = "/var/lib/systemd/coredump",
    ) -> None:
        """
        ``mode`` values:

        * ``ignore``: all failures are ignored
        * ``warn``: test result category is set to "COREDUMPS" and the test is
          marked as such in a test summary, however test outcome and pytest exit
          code is kept intact
        * ``fail``: test result category is set to "COREDUMPS" and the test is
          marked as such in a test summary, if a test outcome is ``passed`` it
          is set to ``failed`` and pytest will return non-zero exit code

        :param host: Multihost host.
        :type host: MultihostHost
        :param mode: Action taken when a core file is found.
        :type mode: Literal["fail", "warn", "ignore"]
        :param filter: Regular expression used to filter the core file names,
            defaults to None
        :type filter: str | None, optional
        :param path: Path to the directory where core files are stored, defaults
            to ``/var/lib/systemd/coredump``
        :type path: str
        """
        super().__init__(host)

        self.fs: LinuxFileSystem = fs

        self.mode: Literal["fail", "warn", "ignore"] = mode
        """Action taken when a core file is found."""

        self.filter: str | None = None
        """Regular expression to filter core file names."""

        self.path: str = path
        """Path to the directory where core files are stored."""

        self._corefiles: list[str] | None = None

    def setup(self) -> None:
        """
        Backup and remove /var/lib/systemd/coredump.

        This directory will be automatically re-created by systemd-coredumpd if
        needed.
        """
        super().setup()

        self.fs.rm("/var/lib/systemd/coredump")

    def list_core_files(self) -> list[str]:
        """
        List available core files.

        :return: List of core file names.
        :rtype: list[str]
        """
        # List the folder, exit with 0 if it does not exist (no core files were produced)
        result = self.host.conn.run(
            f"[ -d '{self.path}' ] && ls -N -1 '{self.path}' || :", log_level=ProcessLogLevel.Error
        )

        return result.stdout_lines

    def parse_core_file_name(self, name: str) -> tuple[str, str]:
        """
        Parse core file name and get the PID and timestamp information.

        Expected format is core.{binary}.{bootid}.{pid}.{timestamp}[.zst]

        :param name: Core file name.
        :type name: str
        :raises ValueError: If the core file name has unexpected format.
        :return: Tuple of ``(PID, timestamp)``.
        :rtype: tuple[str, str] | None
        """
        parts = name.split(".")
        if not parts or len(parts) < 3:
            raise ValueError(f"Invalid coredump name: {name}")

        # If the file is compressed, remove the extension
        if not parts[-1].isnumeric():
            parts.pop()

        pid = parts[-2]
        timestamp = parts[-1]

        return (pid, timestamp)

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Dump backtrace and other information from generated core files for easy access.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        if self._corefiles is None:
            self._corefiles = self.list_core_files()

        if not self._corefiles:
            return set()

        # Parse PID and timestamp that we can use to get information for journal
        for name in self._corefiles:
            try:
                pid, timestamp = self.parse_core_file_name(name)
            except ValueError:
                self.logger.warn(f"Invalid core file name: {name}")
                continue

            # Dump the information
            self.host.conn.run(
                rf"""
                journalctl --output=verbose          \
                    'COREDUMP_PID={pid}'             \
                    'COREDUMP_TIMESTAMP={timestamp}' \
                    > '{self.path}/{name}.backtrace'
                """,
                log_level=ProcessLogLevel.Error,
            )

        return {self.path}

    def pytest_report_teststatus(
        self, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
    ) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
        """
        Report core file found error if found and matches requested filter.

        :param report: Pytest report
        :type report: pytest.CollectReport | pytest.TestReport
        :param config: Pytest config
        :type config: pytest.Config
        :return: Pytest test status
        :rtype: tuple[str, str, str | tuple[str, dict[str, bool]]] | None
        """
        if report.when != "call":
            return None

        if self.mode == "ignore" or report.outcome == "skipped":
            return None

        self.logger.info("Checking for core files")

        if self._corefiles is None:
            self._corefiles = self.list_core_files()

        if not self._corefiles:
            return None

        # Ignore if no core files matches the filter
        if self.filter:
            match = re.search(self.filter, "\n".join(self._corefiles))
            if match is None:
                return None

        original_outcome = report.outcome

        # Fail the test if fail mode is selected
        if report.outcome == "passed" and self.mode == "fail":
            report.outcome = "failed"

        # Count this test into "COREDUMPS" category in the final summary,
        # mark it with "C"/"COREDUMP" in short/verbose listing.
        return ("COREDUMPS", "C", f"{original_outcome.upper()}/COREDUMP")



from __future__ import annotations

import signal
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Self

import colorama as c
from pylibsshext.channel import Channel as LibsshChannel
from pylibsshext.errors import LibsshSessionException
from pylibsshext.session import Session as LibsshSession

from pytest_mh.conn import Process, ProcessLogLevel

from .._private.logging import MultihostLogger
from . import Connection, ConnectionError, ProcessError, ProcessInputBuffer, ProcessResult, ProcessTimeoutError, Shell

if TYPE_CHECKING:
    from .. import MultihostHost

__all__ = [
    "SSHClient",
    "SSHAuthenticationError",
    "SSHProcess",
    "SSHProcessError",
    "SSHProcessTimeoutError",
    "SSHInputBuffer",
    "SSHProcessResult",
]


class SSHInputBuffer(ProcessInputBuffer):
    """
    SSH Input Buffer.

    Allows to write into stdin of opened SSH channel.
    """

    def __init__(self, channel: LibsshChannel) -> None:
        """
        :param channel: Opened libssh channel.
        :type channel: LibsshChannel
        """
        self._channel: LibsshChannel = channel

    def write(self, data: str | bytes) -> None:
        if isinstance(data, str):
            data = data.encode("utf-8")

        self._channel.write(data)


class SSHOutputBuffer(Generator):
    """
    SSH Output Buffer.

    Reads from stdout or stderr from an opened SSH channel and makes each line
    of the data accessible through a generator.
    """

    def __init__(self, channel: LibsshChannel, stderr: bool):
        """
        :param channel: Opened libssh channel.
        :type channel: LibsshChannell
        :param stderr: Whether to read stdout or stderr.
        :type stderr: bool
        """
        self.channel: LibsshChannel = channel
        self.stderr: bool = stderr

        self.eof: bool = False
        self.chunk: str = ""
        self.lines: list[str] = []

    def _read(self) -> str:
        """
        Read available data.

        :rtype: str
        """
        chunk: bytes = b""
        while not self.eof:
            self.channel.poll(timeout=1000, stderr=self.stderr)
            new_chunk: bytes | None = self.channel.recv(stderr=self.stderr)

            if new_chunk is None:
                self.eof = True
            else:
                chunk += new_chunk

            try:
                out = chunk.decode("utf-8")
                return out
            except UnicodeDecodeError:
                # Error if we don't have anything more to read
                if self.eof:
                    raise

                # Otherwise concat with next chunk and see if we can decode it then
                continue

        return ""

    def finish(self) -> None:
        """
        Read all remaining data.
        """
        list(self)

    def read_once_into_buffer(self) -> None:
        """
        Read all data that are currently available and store it in the lines
        buffer.
        """
        self.chunk += self._read()

        while self.chunk:
            newline = self.chunk.find("\n")

            # Store a full line if it was already read
            if newline >= 0:
                line = self.chunk[:newline]
                self.lines.append(line)
                self.chunk = self.chunk[newline + 1 :]
                continue

            # Return remaining data if there is nothing else to read
            line = self.chunk
            self.lines.append(line)
            self.chunk = ""

    def send(self, value: Any):
        while True:
            if self.chunk:
                newline = self.chunk.find("\n")

                # Return a full line if it was already read
                if newline >= 0:
                    line = self.chunk[:newline]
                    self.lines.append(line)
                    self.chunk = self.chunk[newline + 1 :]
                    return line

                # Return remaining data if there is nothing else to read
                if self.eof and self.chunk:
                    line = self.chunk
                    self.lines.append(line)
                    self.chunk = ""
                    return line

            # Stop if we can not read anything more
            if self.eof and not self.chunk:
                raise StopIteration

            # Read more data
            self.chunk += self._read()

    def throw(self, typ, val=None, tb=None):
        super().throw(typ, val, tb)


class SSHProcessError(ProcessError):
    """
    SSH Process Error.
    """

    pass


class SSHProcessTimeoutError(ProcessTimeoutError):
    """
    SSH Process Timeout Error.
    """

    pass


class SSHProcessResult(ProcessResult[SSHProcessError]):
    """
    SSH Process result.
    """

    pass


class SSHProcess(Process[SSHProcessResult, SSHInputBuffer, SSHProcessTimeoutError]):
    """
    SSH Process manager.
    """

    def __init__(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        shell: Shell,
        logger: MultihostLogger,
        log_level: ProcessLogLevel,
        timeout: int,
        blocking_call: bool,
        client: SSHClient,
        conn: LibsshSession,
    ) -> None:
        """
        :param command: Command to execute.
        :type command: str
        :param cwd: Working directory, defaults to None
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | bytes | None, optional
        :param shell: Shell used to execute the command, defaults to None (use user's login shell)
        :type shell: str | None, optional
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param log_level: Log level.
        :type log_level: ProcessLogLevel
        :param timeout: Timeout in seconds, value ``0`` means that timeout is
            disabled.
        :type timeout: int
        :param blocking_call: Is this a blocking execution?
        :type blocking_call: bool
        :param client: SSH client.
        :type client: SSHClient
        :param conn: Connected SSH session.
        :type conn: LibsshSession
        """
        super().__init__(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            shell=shell,
            logger=logger,
            log_level=log_level,
            timeout=timeout,
            blocking_call=blocking_call,
            additional_log_data={
                "Host": client.host,
                "User": client.user,
            },
        )

        self.__conn: LibsshSession = conn
        self.__channel: LibsshChannel | None = None

        self.__stdout: SSHOutputBuffer | None = None
        self.__stderr: SSHOutputBuffer | None = None
        self.__stdin: SSHInputBuffer | None = None

    @property
    def in_progress(self) -> bool:
        return self.__channel is not None

    @property
    def stdout(self) -> Generator[str, None, None]:
        if not self.in_progress or self.__stdout is None:
            raise RuntimeError("Accessing stdout on a process that is not running.")

        return self.__stdout

    @property
    def stderr(self) -> Generator[str, None, None]:
        if not self.in_progress or self.__stderr is None:
            raise RuntimeError("Accessing stderr on a process that is not running.")

        return self.__stderr

    @property
    def stdin(self) -> SSHInputBuffer:
        if not self.in_progress or self.__stdin is None:
            raise RuntimeError("Accessing stdin on a process that is not running.")

        return self.__stdin

    def _run(self) -> None:
        """
        Execute the command.

        This is an internal method called by :meth:`run` after executing
        generic code.
        """
        self.__channel = self.__conn.new_channel()
        try:
            self.__channel.request_exec(self.full_command_line)
            self.__stdout = SSHOutputBuffer(self.__channel, stderr=False)
            self.__stderr = SSHOutputBuffer(self.__channel, stderr=True)
            self.__stdin = SSHInputBuffer(self.__channel)

            if self.input is not None:
                self.stdin.write(self.input)
        except Exception:
            self._close()
            raise

    def _wait(self) -> SSHProcessResult:
        """
        Wait for the command to finish.

        EOF is send to standard input to indicate that there will be no
        additional input data. Then it waits for the command to finish.

        This is an internal method called by :meth:`run` after executing
        generic code.

        :return: Command result.
        :rtype: SSHProcessResult
        """
        if self.__stdout is None or self.__stderr is None:
            raise RuntimeError("Calling wait on process that has not yet started.")

        try:
            # Notify the program that there will be no more input
            self.send_eof()

            # Wait for the program to finish and get the exit code.
            code = self._wait_for_rc()

            # Read remaining output, this will finish the output generator and append
            # remaining lines to self.__stdout and self.__stderr buffers.
            self.__stdout.finish()
            self.__stderr.finish()

            error = SSHProcessError(
                code, self.id, self.command, self.cwd, self.env, self.input, self.__stdout.lines, self.__stderr.lines
            )

            result = SSHProcessResult(code, self.__stdout.lines, self.__stderr.lines, error)
        except TimeoutError as e:
            self.__stdout.read_once_into_buffer()
            self.__stderr.read_once_into_buffer()
            raise SSHProcessTimeoutError(
                e.args[0],
                self.id,
                self.command,
                self.cwd,
                self.env,
                self.input,
                self.__stdout.lines,
                self.__stderr.lines,
            )
        finally:
            self._close()

        return result

    def send_eof(self) -> None:
        if self.__channel is None:
            raise RuntimeError("Calling send_eof on process that is not running.")

        self.__channel.send_eof()

    def send_signal(self, sig: signal.Signals) -> None:
        if self.__channel is None:
            raise RuntimeError("Calling send_signal on process that is not running.")

        self.__channel.send_signal(sig)

    def _close(self) -> None:
        if self.__channel is None:
            return

        self.__channel.close()
        self.__channel = None
        self.__stdout = None
        self.__stderr = None
        self.__stdin = None

    def _wait_for_rc(self) -> int:
        if self.__channel is None:
            raise RuntimeError("Calling _wait_for_rc on process that is not running.")

        rc = -1
        while rc == -1:
            rc = self.__channel.get_channel_exit_status()
            if rc == -1:
                self.__channel.poll(timeout=5)

        return rc


class SSHAuthenticationError(ConnectionError):
    """
    Unable to authenticate over SSH.
    """

    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        message: str,
    ) -> None:
        super().__init__(f'Unable to authenticate as "{user}" at {host}:{port} over SSH: {message}')


class SSHClient(Connection[SSHProcess, SSHProcessResult]):
    """
    Interactive SSH client.
    """

    def __init__(
        self,
        host: str,
        *,
        user: str,
        password: str | None = None,
        private_key_path: str | Path | None = None,
        private_key_password: str | None = None,
        port: int = 22,
        shell: Shell,
        logger: MultihostLogger,
        timeout: int = 300,
    ) -> None:
        """
        :param host: Host name to connect to.
        :type host: str
        :param user: Username to authenticate.
        :type user: str
        :param password: Password for authentication, defaults to ``None``.
        :type password: str | None
        :param private_key_path: Path to the private key for authentication, defaults to ``None``.
        :type private_key_path: str | Path | None
        :param private_key_password: Password to unlock the private key, defaults to ``None``.
        :type private_key_password: str | None
        :param port: SSH port, defaults to 22
        :type port: int, optional
        :param shell: User shell used to run commands, defaults to SSHBashProcess
        :type shell: str, optional
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param timeout: Timeout in seconds (defaults to 300), value
            ``0`` means that timeout is disabled.
        :type timeout: int
        """
        super().__init__(shell=shell, logger=logger, timeout=timeout)

        if password is None and private_key_path is None:
            raise ValueError("At least one authentication mechanism has to be set.")

        self.host: str = host
        self.user: str = user
        self.port: int = port

        self.password: str | None = password
        self.private_key: bytes | None = None
        self.private_key_password: bytes | None = None
        self.private_key, self.private_key_password = self._read_private_key(private_key_path, private_key_password)

        # Timeout is maximum number of seconds that operation can block, after
        # then it returns error and we need to retry. It is necessary to set,
        # since Python will not deliver signal if the code is blocked in C
        # library. The signal is deliver only after we get back to the Python
        # code.
        self.__conn: LibsshSession = LibsshSession()

    def _read_private_key(
        self,
        path: str | Path | None = None,
        password: str | None = None,
    ) -> tuple[bytes | None, bytes | None]:
        private_key: bytes | None = None
        private_key_password: bytes | None = None

        if path is not None:
            with open(path, "rb") as f:
                private_key = f.read()

        if password is not None:
            private_key_password = password.encode("utf-8")

        return private_key, private_key_password

    @property
    def requested_auth_methods(self) -> list[str]:
        """
        :return: List of authentication methods requested by the caller.
        :rtype: list[str]
        """
        methods = []

        if self.password is not None:
            methods.append("password")

        if self.private_key is not None:
            methods.append("private key")

        return methods

    @property
    def connected(self) -> bool:
        return bool(self.__conn.is_connected)

    def connect(self) -> None:
        """
        Connect to the host.

        :raises SSHAuthenticationError: If user fails to authenticate.
        """
        if self.connected:
            return

        self.logger.info(
            self.logger.colorize("Opening SSH connection to ", c.Style.BRIGHT)
            + self.logger.colorize(self.host, c.Fore.BLUE, c.Style.BRIGHT)
            + self.logger.colorize(f" using {'/'.join(self.requested_auth_methods)}", c.Style.BRIGHT)
        )

        try:
            self.__conn.connect(
                host=self.host,
                user=self.user,
                password=self.password,
                private_key=self.private_key,
                private_key_password=self.private_key_password,
                port=self.port,
                host_key_checking=False,
                open_session_retries=10,
            )
            self.__conn.set_ssh_options("timeout", 1)
        except LibsshSessionException as e:
            raise SSHAuthenticationError(self.host, self.port, self.user, e.message)

    def disconnect(self) -> None:
        self.logger.info(
            self.logger.colorize("Closing SSH connection to ", c.Style.BRIGHT)
            + self.logger.colorize(self.host, c.Fore.BLUE, c.Style.BRIGHT)
        )

        self.__conn.disconnect()

    def create_process(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        log_level: ProcessLogLevel,
        blocking_call: bool,
    ) -> SSHProcess:
        return SSHProcess(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            shell=self.shell,
            logger=self.logger,
            log_level=log_level,
            timeout=self.timeout,
            blocking_call=blocking_call,
            client=self,
            conn=self.__conn,
        )

    @classmethod
    def from_confdict(cls, host: MultihostHost, confdict: dict[str, Any]) -> Self:
        ssh_host: str = confdict.get("host", host.hostname)
        ssh_port: int = int(confdict.get("port", 22))
        ssh_username: str = confdict.get("username", "root")
        ssh_password: str | None = confdict.get("password", None)
        ssh_private_key: str | None = confdict.get("private_key", None)
        ssh_private_key_password: str | None = confdict.get("private_key_password", None)
        timeout: int = confdict.get("timeout", 300)

        if ssh_password is None and ssh_private_key is None:
            ssh_password = "Secret123"

        return cls(
            host=ssh_host,
            user=ssh_username,
            password=ssh_password,
            private_key_path=ssh_private_key,
            private_key_password=ssh_private_key_password,
            port=ssh_port,
            logger=host.logger,
            shell=host.shell,
            timeout=timeout,
        )



from __future__ import annotations

import signal
from collections.abc import Mapping
from copy import deepcopy
from functools import partial, wraps
from inspect import getfullargspec
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, ParamSpec, TypeVar

from .types import MultihostOutcome

if TYPE_CHECKING:
    from .artifacts import MultihostArtifactsMode


Param = ParamSpec("Param")
RetType = TypeVar("RetType")


def timeout(
    seconds: int, message: str = "Operation timed out"
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """
    Raise TimeoutError if function takes longer then ``seconds`` to finish.

    :param seconds: Number of seconds to wait.
    :type seconds: int
    :param message: Exception message, defaults to "Operation timed out"
    :type message: str, optional
    :raises ValueError: If ``seconds`` is less or equal to zero.
    :raises TimeoutError: If timeout occurrs.
    :return: Decorator.
    :rtype: Callable[[Callable[Param, RetType]], Callable[Param, RetType]]
    """
    if seconds < 0:
        raise ValueError(f"Invalid timeout value: {seconds}")

    def _timeout_handler(signum, frame):
        raise TimeoutError(seconds, message)

    def decorator(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
        if seconds == 0:
            return func

        @wraps(func)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
            old_handler = signal.signal(signal.SIGALRM, _timeout_handler)
            old_timer = signal.setitimer(signal.ITIMER_REAL, seconds)
            try:
                return func(*args, **kwargs)
            finally:
                signal.setitimer(signal.ITIMER_REAL, *old_timer)
                signal.signal(signal.SIGALRM, old_handler)

        return wrapper

    return decorator


def validate_configuration(
    required_keys: list[str], confdict: dict[str, Any], error_fmt: str = '"{key}" property is missing'
) -> None:
    """
    Validate configuration dictionary.

    Check that it contains all required keys. The key may contain ``.`` to check
    nested keys, for example ``ssh.user``.

    :param required_keys: Required keys.
    :type required_keys: list[str]
    :param confdict: Configuration dictionary.
    :type confdict: dict[str, Any]
    :param error_fmt: _description_, defaults to '"{key}" property is missing'
    :type error_fmt: str, optional
    :raises ValueError: If a required key is missing or empty.
    :return: ``True`` if all keys are present and not empty, ``False`` otherwise.
    :rtype: bool
    """

    def is_property_in_dict(property: str, d: dict[str, Any]) -> bool:
        if "." in property:
            (key, subpath) = property.split(".", maxsplit=1)
            if not d.get(key, None):
                return False

            return is_property_in_dict(subpath, d[key])

        return isinstance(d, Mapping) and property in d and d[property]

    for key in required_keys:
        if not is_property_in_dict(key, confdict):
            raise ValueError(error_fmt.format(key=key))


def merge_dict(*args: dict | None):
    """
    Merge two or more nested dictionaries together.

    Nested dictionaries are not overwritten but combined.
    """
    filtered_args = [x for x in args if x is not None]
    if not filtered_args:
        return {}

    dest = deepcopy(filtered_args[0])

    for source in filtered_args[1:]:
        for key, value in source.items():
            if isinstance(value, dict):
                dest[key] = merge_dict(dest.get(key, {}), value)
                continue

            dest[key] = value

    return dest


def invoke_callback(cb: Callable, /, **kwargs: Any) -> Any:
    """
    Invoke callback with given arguments.

    The callback may take all or only selected subset of given arguments.

    :param cb: Callback to call.
    :type cb: Callable
    :param `*kwargs`: Callback parameters.
    :type `*kwargs`: dict[str, Any]
    :param
    :return: Return value of the callabck.
    :rtype: Any
    """
    cb_args: list[str] = []

    # Get list of parameters required by the callback
    if isinstance(cb, partial):
        spec = getfullargspec(cb.func)

        cb_args = spec.args + spec.kwonlyargs

        # Remove bound positional parameters
        cb_args = cb_args[len(cb.args) :]

        # Remove bound keyword parameters
        cb_args = [x for x in cb_args if x not in cb.keywords]
    else:
        spec = getfullargspec(cb)
        cb_args = spec.args + spec.kwonlyargs

    if spec.varkw is None:
        # No **kwargs is present, just pick selected arguments
        callspec = {k: v for k, v in kwargs.items() if k in cb_args}
    else:
        # **kwargs is present, pass everything
        callspec = kwargs

    return cb(**callspec)


def sanitize_path(path: str | Path) -> Path:
    """
    Replace problematic characters in file path.

    :param path: Path to sanitize.
    :type path: str | Path
    :return: Sanitized path.
    :rtype: Path
    """
    table = str.maketrans('":<>|*? [', "---------", "]()")
    return Path(str(path).translate(table))


def should_collect_artifacts(mode: MultihostArtifactsMode, outcome: MultihostOutcome) -> bool:
    """
    Match mode and outcome in order to decide if artifacts should be
    collected/written or not.

    :param mode: Artifacts collect mode.
    :type mode: MultihostArtifactsMode
    :param outcome: Test or operation outcome.
    :type outcome: MultihostOutcome
    :raises ValueError: If mode is not recognized.
    :return: True if artifacts should be collected, False otherwise.
    :rtype: bool
    """
    match mode:
        case "never":
            return False
        case "always":
            return True
        case "on-failure":
            return outcome in ("failed", "error", "unknown")

    raise ValueError(f"Unexpected artifacts collection mode: {mode}")


class OperationStatus(object):
    """
    Keep named states of operation.

    This can be used to mark certain operations as completed with success,
    failure or any other state and act later upon it.
    """

    def __init__(self) -> None:
        self.states: dict[str, str] = {}

    def set(self, name: str, state: str) -> None:
        """
        Set current state of the operation.

        :param name: Operation name.
        :type name: str
        :param state: Current state.
        :type state: str
        """
        self.states[name] = state

    def clear(self, name: str) -> None:
        """
        Clear state of the operation.

        :param name: Operation name.
        :type name: str
        """
        if name in self.states:
            del self.states[name]

    def set_success(self, name: str) -> None:
        """
        Mark operation as successful.

        :param name: Operation name.
        :type name: str
        """
        self.set(name, "success")

    def set_failure(self, name: str) -> None:
        """
        Mark operation as failed.

        :param name: Operation name.
        :type name: str
        """
        self.set(name, "failure")

    def check(self, name: str, expected_state: str) -> bool:
        """
        Check operation state.

        :param name: Operation name.
        :type name: str
        :param expected_state: Expected state.
        :type expected_state: str
        :return: True if current state equals to the expected state, False otherwise.
        :rtype: bool
        """
        state = self.states.get(name, None)
        if state is None:
            return False

        return state == expected_state

    def check_success(self, name: str) -> bool:
        """
        Check if operation state is success.

        :param name: Operation name.
        :type name: str
        :return: True if current state equals to ``success``, False otherwise.
        :rtype: bool
        """
        return self.check(name, "success")

    def check_failure(self, name: str) -> bool:
        """
        Check if operation state is failure.

        :param name: Operation name.
        :type name: str
        :return: True if current state equals to ``failure``, False otherwise.
        :rtype: bool
        """
        return self.check(name, "failure")



from __future__ import annotations

import pytest

from .marks import TopologyMark
from .multihost import MultihostConfig
from .types import MultihostOutcome


class MultihostItemData(object):
    """
    Multihost internal pytest data, stored in :attr:`pytest.Item.multihost`
    """

    def __init__(self, multihost: MultihostConfig | None, topology_mark: TopologyMark | None) -> None:
        self.multihost: MultihostConfig | None = multihost
        """
        Multihost object.
        """

        self.topology_mark: TopologyMark | None = topology_mark
        """
        Topology mark for the test run.
        """

        self.outcome: MultihostOutcome = "unknown"
        """
        Test run outcome, available in fixture finalizers.
        """

        self.result: pytest.TestReport | None = None
        """
        Pytest test result.
        """

    def _init(self) -> None:
        """
        Postponed initialization. This is called once we know that current
        mh configuration supports desired topology.
        """
        # Initialize topology controller
        if self.multihost is not None and self.topology_mark is not None:
            self.topology_mark.controller.init(
                self.topology_mark.name,
                self.multihost,
                self.multihost.logger,
                self.topology_mark.topology,
                self.topology_mark.fixtures,
            )

    @staticmethod
    def SetData(item: pytest.Item, data: MultihostItemData | None) -> None:
        item.stash[DataStashKey] = data

    @staticmethod
    def GetData(item: pytest.Item) -> MultihostItemData | None:
        return item.stash.get(DataStashKey, None)


DataStashKey = pytest.StashKey[MultihostItemData | None]()



from __future__ import annotations

import re
from typing import Literal

import pytest

from .. import MultihostHost, MultihostUtility
from ..conn import ProcessLogLevel

__all__ = ["Auditd"]


class Auditd(MultihostUtility[MultihostHost]):
    """
    Auditd utilities.

    Collects audit logs and detects AVC denials.
    """

    def __init__(
        self,
        host: MultihostHost,
        *,
        avc_mode: Literal["fail", "warn", "ignore"],
        avc_filter: str | None = None,
    ) -> None:
        """
        ``avc_mode`` values:

        * ``ignore``: all failures are ignored
        * ``warn``: test result category is set to "AVC DENIALS" and the test is
          marked as such in a test summary, however test outcome and pytest exit
          code is kept intact
        * ``fail``: test result category is set to "AVC DENIALS" and the test is
          marked as such in a test summary, if a test outcome is ``passed`` it
          is set to ``failed`` and pytest will return non-zero exit code

        :param host: Multihost host.
        :type host: MultihostHost
        :param avc_mode: Action taken when AVC denial is found in audit logs.
        :type avc_mode: Literal["fail", "warn", "ignore"]
        :param avc_filter: Regular expression used to filter the AVC denials,
            defaults to None
        :type avc_filter: str | None, optional
        """
        super().__init__(host)

        self.avc_mode: Literal["fail", "warn", "ignore"] = avc_mode
        self.avc_filter: str | None = avc_filter

        self.artifacts: set[str] = {"/var/log/audit/audit.log"}
        self._backup: str | None = None
        self._auditd_running: bool = False

    def setup(self) -> None:
        """
        Create backup of audit logs and clear them for current test run.
        """
        super().setup()

        result = self.host.conn.run(
            """
            set -e

            if [ ! -d /var/log/audit ]; then
                exit 0
            fi

            tmp=`mktemp -d`
            cp -r --archive /var/log/audit "$tmp"
            truncate --size 0 /var/log/audit/audit.log*
            echo $tmp
            """,
            log_level=ProcessLogLevel.Error,
        )

        tmp_path = result.stdout.strip()
        if tmp_path:
            self._auditd_running = True
            self._backup = tmp_path

    def teardown(self) -> None:
        """
        Restore previous audit logs from backup and remove the backup.
        """
        if self._backup is not None:
            self.host.conn.run(
                f"""
                set -e

                for f in "{self._backup}"/audit/audit.log*; do
                    name=`basename "$f"`
                    cat "$f" > "/var/log/audit/$name"
                done

                rm -fr "{self._backup}"
                """,
                log_level=ProcessLogLevel.Error,
            )

        return super().teardown()

    def pytest_report_teststatus(
        self, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
    ) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
        """
        Report AVC denial error if found and matches requested filter.

        :param report: Pytest report
        :type report: pytest.CollectReport | pytest.TestReport
        :param config: Pytest config
        :type config: pytest.Config
        :return: Pytest test status
        :rtype: tuple[str, str, str | tuple[str, dict[str, bool]]] | None
        """
        if report.when != "call":
            return None

        if not self._auditd_running or self.avc_mode == "ignore" or report.outcome == "skipped":
            return None

        self.logger.info("Checking for AVC denials")

        result = self.host.conn.run(
            "ausearch --input-logs -m AVC,USER_AVC", raise_on_error=False, log_level=ProcessLogLevel.Silent
        )
        if result.rc:
            return None

        records = result.stdout
        if not records:
            return None

        # Ignore if no message matches the filter
        if self.avc_filter:
            match = re.search(self.avc_filter, records)
            if match is None:
                return None

        original_outcome = report.outcome

        # Fail the test if fail mode is selected
        if report.outcome == "passed" and self.avc_mode == "fail":
            report.outcome = "failed"

        # Count this test into "AVC DENIALS" category in the final summary,
        # mark it with "A"/"AVC DENIAL" in short/verbose listing.
        return ("AVC DENIALS", "A", f"{original_outcome.upper()}/AVC DENIAL")



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from collections import deque
from contextlib import contextmanager
from functools import wraps
from pathlib import Path, PurePath
from typing import TYPE_CHECKING, Any, Generator, Generic, Self, Sequence, Type, TypeVar

import pytest

from ..cli import CLIBuilder
from ..conn import (
    Bash,
    Connection,
    Powershell,
    Process,
    ProcessError,
    ProcessInputBuffer,
    ProcessResult,
    ProcessTimeoutError,
    Shell,
)
from ..conn.container import ContainerClient
from ..conn.ssh import SSHClient
from .artifacts import (
    MultihostArtifactsCollector,
    MultihostArtifactsMode,
    MultihostArtifactsType,
    MultihostHostArtifacts,
)
from .errors import TeardownExceptionGroup
from .logging import MultihostHostLogger, MultihostLogger
from .misc import OperationStatus, validate_configuration
from .topology import Topology
from .types import MultihostOSFamily

if TYPE_CHECKING:
    from .fixtures import MultihostFixture
    from .marks import TopologyMark


class _MultihostDependencyMeta(ABCMeta):
    """
    Base meta class that computes on which utilities a class depends.

    It finds all MultihostUtility of given type and add it to the list of
    dependencies.
    """

    def __call__(cls, types: type[MultihostUtility], *args, **kwargs) -> Any:
        obj = super().__call__(*args, **kwargs)

        # Get list of MultihostUtilities used by this object
        obj._mh_utility_dependencies = []
        deps: list[MultihostUtility] = list()
        for arg in obj.__dict__.values():
            if isinstance(arg, types):
                deps.append(arg)

        # Now sort the dependencies, first by class name, then by cross-utility requirements
        deps = sorted(set(deps), key=lambda x: x.__class__.__name__)

        # Now include utilities that do not depend on other utilities
        for util in deps.copy():
            if not util._mh_utility_dependencies:
                obj._mh_utility_dependencies.append(util)
                deps.remove(util)

        # Now sort all utilities that depends on other
        while deps:
            for util in deps.copy():
                if util._mh_utility_dependencies.issubset(obj._mh_utility_dependencies):
                    obj._mh_utility_dependencies.append(util)
                    deps.remove(util)

        return obj


class _MultihostRoleMeta(_MultihostDependencyMeta):
    """
    MultihostRole metaclass.
    """

    def __call__(cls, *args, **kwargs) -> Any:
        return super().__call__(MultihostUtility, *args, **kwargs)


class _MultihostHostMeta(_MultihostDependencyMeta):
    """
    MultihostHost metaclass.
    """

    def __call__(cls, *args, **kwargs) -> Any:
        return super().__call__(MultihostReentrantUtility, *args, **kwargs)


class _MultihostUtilityMeta(ABCMeta):
    """
    MultihostUtility metaclass.

    It takes care of automatic invocation of setup_when_used.

    ABCMeta is not strictly needed for MultihostUtility, but it is quite
    possible that inherited classed will require ABC, therefore we have to
    support that out of the box.
    """

    def __new__(cls, name: str, bases: tuple, attrs: dict[str, Any]) -> _MultihostUtilityMeta:
        # We only decorate stuff from inherited classes
        if name not in ("MultihostUtility", "MultihostReentrantUtility"):
            for attr, value in attrs.items():
                # Do not decorate private stuff
                if attr.startswith("_"):
                    continue

                # Do not decorate stuff from our base classes
                if attr in MultihostUtility.__dict__ or attr in MultihostReentrantUtility.__dict__:
                    continue

                # Do not decorate @staticmethod and @classmethod
                if isinstance(value, (staticmethod, classmethod)):
                    continue

                # Do not decorate stuff marked with @mh_utility_ignore_use
                if hasattr(value, "_mh_utility_ignore_use") and value._mh_utility_ignore_use:
                    continue

                # Decorate supported fields so when used it will automatically
                # invoke setup_when_used()

                # Methods and other callables
                if callable(value):
                    attrs[attr] = mh_utility_used(value)
                    continue

                # @property
                if isinstance(value, property):
                    fget = mh_utility_used(value.fget) if value.fget is not None else None
                    fset = mh_utility_used(value.fset) if value.fset is not None else None
                    fdel = mh_utility_used(value.fdel) if value.fdel is not None else None

                    attrs[attr] = property(fget, fset, fdel, value.__doc__)
                    continue

        return super().__new__(cls, name, bases, attrs)

    def __init__(self, name: str, bases: tuple, attrs: dict[str, Any]) -> None:
        # define special attributes
        if name in ("MultihostUtility", "MultihostReentrantUtility"):
            self._mh_utility_call_setup = False
            self._mh_utility_call_teardown = False
            self._mh_utility_used = False
            return super().__init__(name, bases, attrs)

        # we only want to call setup and teardown if it is defined in inherited classes
        if "setup" in attrs:
            self._mh_utility_call_setup = True

        if "teardown" in attrs:
            self._mh_utility_call_teardown = True

        super().__init__(name, bases, attrs)


class MultihostConfig(ABC):
    """
    Multihost configuration.
    """

    def __init__(
        self,
        confdict: dict[str, Any],
        *,
        logger: MultihostLogger,
        lazy_ssh: bool,
        artifacts_dir: Path,
        artifacts_mode: MultihostArtifactsMode,
        artifacts_compression: bool,
    ) -> None:
        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in configuration'
        )

        self.confdict: dict[str, Any] = confdict
        """Multihost configuration dictionary given to the constructor."""

        self.config: dict[str, Any] = confdict.get("config", {})
        """Custom configuration."""

        self.logger: MultihostLogger = logger
        """Multihost logger"""

        self.lazy_ssh: bool = lazy_ssh
        """If True, hosts postpone connecting to ssh when the connection is first required"""

        self.artifacts_dir: Path = artifacts_dir
        """Artifacts output directory."""

        self.artifacts_mode: MultihostArtifactsMode = artifacts_mode
        """Artifacts collection mode."""

        self.artifacts_compression: bool = artifacts_compression
        """Store artifacts in compressed archive?"""

        self.domains: list[MultihostDomain] = []
        """Available domains"""

        for domain in confdict["domains"]:
            self.domains.append(self.create_domain(domain))

        self._in_test: bool = False
        """
        Process is currently inside a test.
        """

        self._sigint: bool = False
        """
        SIGINT (CTRL-C) was received.
        """

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the host configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["domains"]

    @property
    def TopologyMarkClass(self) -> Type[TopologyMark]:
        """
        Class name of the type or subtype of :class:`TopologyMark`.
        """
        from .marks import TopologyMark

        return TopologyMark

    def create_domain(self, domain: dict[str, Any]) -> MultihostDomain:
        """
        Create new multihost domain from dictionary.

        It maps the role name to a Python class using
        :attr:`id_to_domain_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it raises
        ``ValueError``.

        :param domain: Domain in dictionary form.
        :type domain: dict[str, Any]
        :raises ValueError: If domain does not have id or mapping to Python class is not found.
        :return: New multihost domain.
        :rtype: MultihostDomain
        """
        id = domain.get("id", None)
        if id is None:
            raise ValueError("Invalid configuration, domain is missing 'id'")

        cls = self.id_to_domain_class.get(id, self.id_to_domain_class.get("*", None))
        if cls is None:
            raise ValueError(f"Unexpected domain id: {id}")

        return cls(self, domain)

    def topology_hosts(self, topology: Topology) -> list[MultihostHost]:
        """
        Return all hosts required by the topology as list.

        :param topology: Topology.
        :type topology: Multihost topology
        :return: List of MultihostHost.
        :rtype: list[MultihostHost]
        """
        result: list[MultihostHost] = []

        for mh_domain in self.domains:
            if mh_domain.id in topology:
                topology_domain = topology.get(mh_domain.id)
                for role_name in mh_domain.roles:
                    if role_name not in topology_domain:
                        continue

                    count = topology_domain.get(role_name)
                    result.extend(mh_domain.hosts_by_role(role_name)[:count])

        return result

    @property
    @abstractmethod
    def id_to_domain_class(self) -> dict[str, Type[MultihostDomain]]:
        """
        Map domain id to domain class. Asterisk ``*`` can be used as fallback
        value.

        :rtype: Class name.
        """
        pass


ConfigType = TypeVar("ConfigType", bound=MultihostConfig)


class MultihostDomain(ABC, Generic[ConfigType]):
    """
    Multihost domain class.
    """

    def __init__(self, config: ConfigType, confdict: dict[str, Any]) -> None:
        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in domain configuration'
        )

        self.confdict: dict[str, Any] = confdict
        """Multihost domain configuration dictionary given to the constructor."""

        self.config: dict[str, Any] = confdict.get("config", {})
        """Custom configuration."""

        self.mh_config: ConfigType = config
        """Multihost configuration"""

        self.logger: MultihostLogger = config.logger
        """Multihost logger"""

        self.id: str = confdict["id"]
        """Domain id"""

        self.hosts: list[MultihostHost] = []
        """Available hosts in this domain"""

        for host in confdict["hosts"]:
            self.hosts.append(self.create_host(host))

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the domain configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["id", "hosts"]

    @property
    def roles(self) -> list[str]:
        """
        All roles available in this domain.

        :return: Role names.
        :rtype: list[str]
        """
        return sorted(set(x.role for x in self.hosts))

    def create_host(self, confdict: dict[str, Any]) -> MultihostHost:
        """
        Create host object from role.

        It maps the role name to a Python class using
        :attr:`role_to_host_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it fallbacks to
        :class:`MultiHost`.

        :param confdict: Host configuration as a dictionary.
        :type confdict: dict[str, Any]
        :raises ValueError: If role property is missing in the host
            configuration.
        :return: Host instance.
        :rtype: MultihostHost
        """
        if not confdict.get("role", None):
            raise ValueError('"role" property is missing in host configuration')

        role = confdict["role"]
        cls = self.role_to_host_class.get(role, self.role_to_host_class.get("*", MultihostHost))

        return cls(self, confdict)

    def create_role(self, mh: MultihostFixture, host: MultihostHost) -> MultihostRole:
        """
        Create role object from given host.

        It maps the role name to a Python class using
        :attr:`role_to_role_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it raises
        ``ValueError``.

        :param mh: Multihost instance.
        :type mh: Multihost
        :param host: Multihost host instance.
        :type host: MultihostHost
        :raises ValueError: If unexpected role name is given.
        :return: Role instance.
        :rtype: MultihostRole
        """
        cls = self.role_to_role_class.get(host.role, self.role_to_role_class.get("*", None))
        if cls is None:
            raise ValueError(f"Unexpected role: {host.role}")

        return cls(mh, host.role, host)

    @property
    @abstractmethod
    def role_to_host_class(self) -> dict[str, Type[MultihostHost]]:
        """
        Map role to host class. Asterisk ``*`` can be used as fallback value.

        :rtype: Class name.
        """
        pass

    @property
    @abstractmethod
    def role_to_role_class(self) -> dict[str, Type[MultihostRole]]:
        """
        Map role to role class. Asterisk ``*`` can be used as fallback value.

        :rtype: Class name.
        """
        pass

    def hosts_by_role(self, role: str) -> list[MultihostHost]:
        """
        Return all hosts of the given role.

        :param role: Role name.
        :type role: str
        :return: List of hosts of given role.
        :rtype: list[MultihostHost]
        """
        return [x for x in self.hosts if x.role == role]


DomainType = TypeVar("DomainType", bound=MultihostDomain)


class MultihostHost(Generic[DomainType], metaclass=_MultihostHostMeta):
    """
    Base multihost host class.

    .. note::

        Host objects may contain MultihostReentrantUtility objects. These
        utilities are automatically setup, entered, exited and teared down.

        It may also contain MultihostUtility objects, but setup and teardown of
        these utilities must be handled manually by in the host or topology
        setup/teardown methods to create the required scope.

    .. code-block:: yaml
        :caption: Example configuration in YAML format

        - hostname: dc.ad.test
          role: ad
          os:
            family: linux
          ssh:
            host: 1.2.3.4
            username: root
            password: Secret123
          config:
            binddn: Administrator@ad.test
            bindpw: vagrant
            client:
              ad_domain: ad.test
              krb5_keytab: /enrollment/ad.keytab
              ldap_krb5_keytab: /enrollment/ad.keytab

    * Required fields: ``hostname``, ``role``
    * Optional fields: ``artifacts``, ``config``, ``os``, ``ssh``
    """

    # Following attributes are set by metaclass
    _mh_utility_dependencies: list[MultihostUtility]

    def __init__(self, domain: DomainType, confdict: dict[str, Any]):
        """
        :param domain: Multihost domain object.
        :type domain: DomainType
        :param confdict: Host configuration as a dictionary.
        :type confdict: dict[str, Any]
        """
        self._op_state: OperationStatus = OperationStatus()
        """Keep state of setup and teardown methods."""

        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in host configuration'
        )

        self.confdict: dict[str, Any] = confdict
        """Multihost host configuration dictionary given to the constructor."""

        # Required
        self.mh_domain: DomainType = domain
        """Multihost domain."""

        self.role: str = confdict["role"]
        """Host role."""

        self.hostname: str = confdict["hostname"]
        """Host hostname."""

        self.logger: MultihostLogger = self.mh_domain.logger.subclass(
            cls=MultihostHostLogger, suffix=f"host.{self.hostname}", hostname=self.hostname
        )
        """Multihost logger."""

        # Optional
        self.config: dict[str, Any] = confdict.get("config", {})
        """Custom configuration."""

        self.configured_artifacts: MultihostHostArtifacts = MultihostHostArtifacts(confdict.get("artifacts", []))
        """Host artifacts produced during tests, configured by the user."""

        # Get host operating system information
        os = confdict.get("os", {})

        os_family = str(os.get("family", MultihostOSFamily.Linux.value)).lower()
        try:
            self.os_family: MultihostOSFamily = MultihostOSFamily(os_family)
            """Host operating system os_family."""
        except ValueError:
            raise ValueError(f'Value "{os_family}" is not supported in os_family field of host configuration')

        # Not configurable, since we expect specific shells in our code
        self.shell: Shell = Bash()
        """Shell used to run commands over host connection."""

        # Set host shell based on the operating system
        match self.os_family:
            case MultihostOSFamily.Linux:
                pass
            case MultihostOSFamily.Windows:
                self.shell = Powershell()
            case _:
                raise ValueError(f"Unknown operating system os_family: {self.os_family}")

        # Connection to the host
        self.conn: Connection[
            Process[ProcessResult, ProcessInputBuffer, ProcessTimeoutError], ProcessResult[ProcessError]
        ] = self.get_connection()
        """Connection to the host."""

        # CLI Builder instance
        self.cli: CLIBuilder = CLIBuilder(self.shell)
        """Command line builder."""

        self.artifacts: MultihostHostArtifacts = MultihostHostArtifacts()
        """
        List of artifacts that will be automatically collected at specific
        places. This list can be dynamically extended. Values may contain
        wildcard character.
        """

        self.artifacts_collector: MultihostArtifactsCollector = MultihostArtifactsCollector(
            host=self,
            path=self.mh_domain.mh_config.artifacts_dir,
            mode=self.mh_domain.mh_config.artifacts_mode,
            compress=self.mh_domain.mh_config.artifacts_compression,
        )
        """
        Artifacts collector.
        """

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the host configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["role", "hostname"]

    def pytest_setup(self) -> None:
        """
        Called once before execution of any tests.
        """
        pass

    def pytest_teardown(self) -> None:
        """
        Called once after all tests are finished.
        """
        pass

    def setup(self) -> None:
        """
        Called before execution of each test.
        """
        pass

    def teardown(self) -> None:
        """
        Called after execution of each test.
        """
        pass

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Return the list of artifacts to collect.

        This just returns :attr:`artifacts`, but it is possible to override this
        method in order to generate additional artifacts that were not created
        by the test, or detect which artifacts were created and update the
        artifacts list.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        return self.configured_artifacts.get(artifacts_type) | self.artifacts.get(artifacts_type)

    def get_connection(self) -> Connection:
        """
        Get connection object to the host with given shell.

        This creates a connection object using the information from the
        multihost configuration. The caller should not make any assumptions
        about the connection mechanism.

        :return: Generic connection to the host.
        :rtype: Connection
        """
        conn_confdict = self.confdict.get("conn", {})
        conn_type = conn_confdict.setdefault("type", "ssh")

        match conn_type:
            case "ssh":
                return SSHClient.from_confdict(self, conn_confdict)
            case "podman" | "docker":
                return ContainerClient.from_confdict(self, conn_confdict)
            case _:
                raise ValueError(f"Unknown connection type: {conn_type}!")


class MultihostBackupHost(MultihostHost[DomainType], ABC):
    """
    Abstract class implementing automatic backup and restore for a host.

    A backup of the host is created once when pytest starts and the host is
    restored automatically (unless disabled) when a test run is finished.

    If the backup data is stored as :class:`~pathlib.PurePath` or a sequence of
    :class:`~pathlib.PurePath`, the file is automatically removed from the host
    when all tests are finished. Otherwise no action is done -- it is possible
    to overwrite :meth:`remove_backup` to clean up your data if needed.

    It is required to implement :meth:`start`, :meth:`stop`, :meth:`backup` and
    :meth:`restore`. The :meth:`start` method is called in :meth:`pytest_setup`
    unless ``auto_start`` is set to False and the implementation of this method
    may raise ``NotImplementedError`` which will be ignored.

    By default, the host is reverted when each test run is finished. This may
    not always be desirable and can be disabled via ``auto_restore`` parameter
    of the constructor.
    """

    def __init__(self, *args, auto_start: bool = True, auto_restore: bool = True, **kwargs) -> None:
        """
        :param auto_start: Automatically start service before taking the first
            backup.
        :type auto_restore: bool, optional
        :param auto_restore: If True, the host is automatically restored to the
            backup state when a test is finished in :meth:`teardown`, defaults
            to True
        :type auto_restore: bool, optional
        """
        super().__init__(*args, **kwargs)

        self.backup_data: PurePath | Sequence[PurePath] | Any | None = None
        """Backup data of vanilla state of this host."""

        self._backup_auto_start: bool = auto_start
        """
        If True, the host is automatically started prior taking the first
        backup.
        """

        self._backup_auto_restore: bool = auto_restore
        """
        If True, the host is automatically restored to the backup state when a
        test is finished in :meth:`teardown`.
        """

    def pytest_setup(self) -> None:
        """
        Start the services via :meth:`start` and take a backup by calling
        :meth:`backup`.
        """
        # Make sure required services are running
        if self._backup_auto_start:
            try:
                self.start()
            except NotImplementedError:
                pass

        # Create backup of initial state
        self.backup_data = self.backup()

    def pytest_teardown(self) -> None:
        """
        Remove backup files from the host (calls :meth:`remove_backup`).
        """
        self.remove_backup(self.backup_data)

    def teardown(self) -> None:
        """
        Restore the host from the backup by calling :meth:`restore`.
        """
        if self._backup_auto_restore:
            self.restore(self.backup_data)

        super().teardown()

    def remove_backup(self, backup_data: PurePath | Sequence[PurePath] | Any | None) -> None:
        """
        Remove backup data from the host.

        If backup_data is not :class:`~pathlib.PurePath` or a sequence of
        :class:`~pathlib.PurePath`, this will not have any effect. Otherwise,
        the paths are removed from the host.

        :param backup_data: Backup data.
        :type backup_data: PurePath | Sequence[PurePath] | Any | None
        """
        if backup_data is None:
            return

        if isinstance(backup_data, PurePath):
            backup_data = [backup_data]

        if isinstance(backup_data, Sequence):
            only_paths = True
            for item in backup_data:
                if not isinstance(item, PurePath):
                    only_paths = False
                    break

            if only_paths:
                if isinstance(self.conn.shell, Powershell):
                    for item in backup_data:
                        path = str(item)
                        self.conn.exec(["Remove-Item", "-Force", "-Recurse", path])
                else:
                    for item in backup_data:
                        path = str(item)
                        self.conn.exec(["rm", "-fr", path])

    @abstractmethod
    def start(self) -> None:
        """
        Start required services.

        :raises NotImplementedError: If start operation is not supported.
        """
        pass

    @abstractmethod
    def stop(self) -> None:
        """
        Stop required services.

        :raises NotImplementedError: If stop operation is not supported.
        """
        pass

    @abstractmethod
    def backup(self) -> PurePath | Sequence[PurePath] | Any | None:
        """
        Backup backend data.

        Returns directory or file path where the backup is stored (as
        :class:`~pathlib.PurePath` or sequence of :class:`~pathlib.PurePath`) or
        any Python data relevant for the backup. This data is passed to
        :meth:`restore` which will use this information to restore the host to
        its original state.

        :return: Backup data.
        :rtype: PurePath | Sequence[PurePath] | Any | None
        """
        pass

    @abstractmethod
    def restore(self, backup_data: Any | None) -> None:
        """
        Restore data from the backup.

        :param backup_data: Backup data.
        :type backup_data: PurePath | Sequence[PurePath] | Any | None
        """
        pass


HostType = TypeVar("HostType", bound=MultihostHost)


class MultihostRole(Generic[HostType], metaclass=_MultihostRoleMeta):
    """
    Base role class. Roles are the main interface to the remote hosts that can
    be directly accessed in test cases as fixtures.

    All changes to the remote host that were done through the role object API
    are automatically reverted when a test is finished.

    .. note::

        MultihostRole uses custom metaclass that inherits from ABCMeta.
        Therefore all subclasses can use @abstractmethod any other abc
        decorators without directly inheriting ABCMeta class from ABC.
    """

    # Following attributes are set by metaclass
    _mh_utility_dependencies: list[MultihostUtility]

    def __init__(
        self,
        mh: MultihostFixture,
        role: str,
        host: HostType,
    ) -> None:
        self._op_state: OperationStatus = OperationStatus()
        """Keep state of setup and teardown methods."""

        self.mh: MultihostFixture = mh
        self.role: str = role
        self.host: HostType = host

        self.logger: MultihostLogger = self.host.logger
        """Multihost logger."""

        self.artifacts: set[str] = set()
        """
        List of artifacts that will be automatically collected at specific
        places. This list can be dynamically extended. Values may contain
        wildcard character.
        """

    def setup(self) -> None:
        """
        Called before execution of each test.
        """
        pass

    def teardown(self) -> None:
        """
        Called after execution of each test.
        """
        pass

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Return the list of artifacts to collect.

        This just returns :attr:`artifacts`, but it is possible to override this
        method in order to generate additional artifacts that were not created
        by the test, or detect which artifacts were created and update the
        artifacts list.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        return self.artifacts


class MultihostUtility(Generic[HostType], metaclass=_MultihostUtilityMeta):
    """
    Base class for utility functions that operate on remote hosts, such as
    writing a file or managing SSSD.

    Instances of :class:`MultihostUtility` can be used in any role class which
    is a subclass of :class:`MultihostRole`. In this case, :func:`setup` and
    :func:`teardown` methods are called automatically when the object is created
    and destroyed to ensure proper setup and clean up on the remote host.

    .. note::

        MultihostUtility uses custom metaclass that inherits from ABCMeta.
        Therefore all subclasses can use @abstractmethod any other abc
        decorators without directly inheriting ABCMeta class from ABC.
    """

    # Classvar. This can be overridden by mh_utility_postpone_setup.
    _mh_utility_postpone_setup: bool = False

    # Following attributes are set by metaclass
    _mh_utility_call_setup: bool
    _mh_utility_call_teardown: bool
    _mh_utility_used: bool

    # Set in __new__
    _mh_utility_dependencies: set[MultihostUtility]

    def __new__(cls, *args, **kwargs):
        """
        Find all MultihostUtility objects in the constructor.
        """
        obj = super().__new__(cls)
        obj._mh_utility_dependencies = set()

        for arg in [*args, *kwargs.values()]:
            if isinstance(arg, MultihostUtility):
                obj._mh_utility_dependencies.add(arg)

        return obj

    def __init__(self, host: HostType) -> None:
        self._op_state: OperationStatus = OperationStatus()
        """Keep state of setup and teardown methods."""

        """
        :param host: Remote host instance.
        :type host: HostType
        """
        self.host: HostType = host
        """Multihost host."""

        self.logger: MultihostLogger = self.host.logger
        """Multihost logger."""

        self.artifacts: set[str] = set()
        """
        List of artifacts that will be automatically collected at specific
        places. This list can be dynamically extended. Values may contain
        wildcard character.
        """

    def setup(self) -> None:
        """
        Setup object.
        """
        pass

    def teardown(self) -> None:
        """
        Teardown object.
        """
        pass

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Return the list of artifacts to collect.

        This just returns :attr:`artifacts`, but it is possible to override this
        method in order to generate additional artifacts that were not created
        by the test, or detect which artifacts were created and update the
        artifacts list.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        return self.artifacts

    def postpone_setup(self) -> Self:
        """
        Postpone setup on this instance of MultihostUtility.

        .. code-block:: python
            :caption: Example usage

            class MyRole(MultihostRole):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                    self.firewall: Firewalld = Firewalld(self.host).postpone_setup()

        :return: Self.
        :rtype: Self
        """
        return mh_utility_postpone_setup(self)

    def pytest_report_teststatus(
        self, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
    ) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
        """
        See pytest built-in hook
        :func:`~_pytest.hookspec.pytest_report_teststatus` for more information.

        .. warning::

            This hook is currently called only if ``report.when == 'call'``,
            that is only after the test is run. This may however change in the
            future, therefore it is recommended to add a test to your code as
            well.

            .. code-block:: python

                class Example(MultihostUtility):
                    def pytest_report_teststatus(self, report, config):
                        if report.when != 'call':
                            return None

                        return ("error", "X", "MYERROR")

        :param report: Pytest report.
        :type report: pytest.CollectReport | pytest.TestReport
        :param config: Pytest config.
        :type config: pytest.Config
        :return: Test status.
        :rtype: tuple[str, str, str | tuple[str, Mapping[str, bool]]] | None
        """
        return None


class MultihostReentrantUtility(MultihostUtility[HostType]):
    """
    Reentrant multihost utility.

    It provides the __enter__ and __exit__ abstract methods that can be called
    multiple times in order to create nested states. The implementation of
    __enter__ should save current state and __exit__ should restore hosts into
    this state.

    The utility can be used as a context manager, leaving the context will
    restore the system to the state during the context enter.
    """

    def __init__(self, host: HostType) -> None:
        super().__init__(host)
        self._mh_exit_stack: deque[tuple[str, bool]] = deque()

    @abstractmethod
    def __enter__(self) -> MultihostReentrantUtility:
        """
        Enter new utility context.

        Typically, a utility saves its state and starts a fresh context.

        :return: Self.
        :rtype: MultihostReentrantUtility
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Leave utility context.

        Typically, a utility restores its state to the previous context and
        reverts all changes done on the host.
        """
        pass


def mh_utility_postpone_setup(cls):
    """
    Class decorator that will postpone calling setup of :class:`MultihostUtility`.

    Decorated class will not invoke setup() before each test immediately but it
    will be postponed to the point when the utility is actually used for the
    first time in the test. This can be used to avoid costly utility setup
    on utilities that are used only sporadically.

    If the utility is not used then setup and teardown method are ignored.

    .. code-block:: python
        :caption: Example

        @mh_utility_postpone_setup
        class ExampleUtility(MultihostUtility):
            def setup(self):
                pass

            def teardown(self):
                pass

    .. seealso::

        There are other decorators that can affect the behavior of postponed
        setup.

            * :func:`mh_utility_used`
            * :func:`mh_utility_ignore_use`

    :param cls: Class to decorate.
    :type cls: type
    :return: Decorated class.
    :rtype: type
    """
    cls._mh_utility_postpone_setup = True
    return cls


def mh_utility_used(method):
    """
    Decorator for :class:`MultihostUtility` methods defined in inherited classes.

    Calling decorated method will first invoke :meth:`MultihostUtility.setup`,
    unless other methods were already called.

    .. note::

        Callables and methods decorated with @property are decorated
        automatically.

        This decorator can be used to decorate fields declared in __init__ or
        descriptors not handled by pytest-mh.

    :param method: Method to decorate.
    :type method: Callable
    :return: Decorated method.
    :rtype: Callable
    """

    @wraps(method)
    def wrapper(self: MultihostUtility, *args, **kwargs):
        if self._mh_utility_postpone_setup and not self._mh_utility_used:
            self._mh_utility_used = True
            mh_utility_setup(self)

            if isinstance(self, MultihostReentrantUtility):
                # Last enter was skipped because the utility was not yet used,
                # we will call it now to save state that we can return in exit.
                where, _ = self._mh_exit_stack.pop()
                mh_utility_enter(self, where)

        return method(self, *args, **kwargs)

    return wrapper


def mh_utility_ignore_use(method):
    """
    Decorator for :class:`MultihostUtility` methods defined in inherited classes.

    Decorated method will not count as "using the class" and therefore it
    will not invoke :meth:`MultihostUtility.setup`.

    .. note::

        This is the opposite of :func:`mh_utility_used`.

    :param method: Method to decorate.
    :type method: Callable
    :return: Decorated method.
    :rtype: Callable
    """
    method._mh_utility_ignore_use = True
    return method


def mh_utility_setup(util: MultihostUtility) -> None:
    """
    Setup MultihostUtility.

    :param util: Multihost utility object.
    :type util: MultihostUtility
    """
    if util._mh_utility_postpone_setup and not util._mh_utility_used:
        return

    if not util._mh_utility_call_setup:
        util._op_state.set_success("setup")
        return

    util.setup()
    util._op_state.set_success("setup")


def mh_utility_teardown(util: MultihostUtility) -> None:
    """
    Teardown MultihostUtility.

    :param util: Multihost utility object.
    :type util: MultihostUtility
    """
    if util._mh_utility_postpone_setup and not util._mh_utility_used:
        return

    if not util._mh_utility_call_teardown:
        return

    if not util._op_state.check_success("setup"):
        return

    util.teardown()


def mh_utility_enter(util: MultihostUtility, where: str) -> None:
    """
    Enter MultihostReentrantUtility.

    This is essentially noop if anything else then MultihostReentrantUtility
    is given, but we keep the type broader to simplify code flow.

    :param util: Multihost utility.
    :type util: MultihostUtility
    :param where: Where do we enter the utility.
    :type where: str
    """
    if not isinstance(util, MultihostReentrantUtility):
        return

    # Do not enter the utility if it was not used yet but postpone setup was requested.
    # However we add a mocked record to indicate that exit should not be called either.
    if util._mh_utility_postpone_setup and not util._mh_utility_used:
        util._mh_exit_stack.append((where, False))
        return

    # We cannot enter/exit if setup was not successful
    if not util._op_state.check_success("setup"):
        raise RuntimeError("Trying to call utility enter without successful setup")

    try:
        util.__enter__()
        util._mh_exit_stack.append((where, True))
    except Exception:
        util._mh_exit_stack.append((where, False))
        raise


def mh_utility_exit(util: MultihostUtility, where: str) -> None:
    """
    Exit MultihostReentrantUtility.

    This is essentially noop if anything else then MultihostReentrantUtility
    is given, but we keep the type broader to simplify code flow.

    :param util: Multihost utility.
    :type util: MultihostUtility
    :param where: Where did we enter the utility.
    :type where: str
    """
    if not isinstance(util, MultihostReentrantUtility):
        return

    # We cannot enter/exit if setup was not successful
    if not util._op_state.check_success("setup"):
        return

    if not util._mh_exit_stack:
        raise IndexError("Calling exit but enter was not called")

    enter_where, enter_result = util._mh_exit_stack.pop()
    if enter_where != where:
        util._mh_exit_stack.append((enter_where, enter_result))
        raise IndexError(f"Calling exit from unexpected place {where}, expected {enter_where}")

    if not enter_result:
        # enter failed, let's not do exit
        return

    util.__exit__(None, None, None)


@contextmanager
def mh_utility(util: MultihostUtility) -> Generator[MultihostUtility, None, None]:
    """
    On-demand use of a multihost utility object with a context manager.

    This can be used to automatically setup and teardown a multihost utility
    object that is created on-demand inside a test.

    .. code-block:: python

        with mh_utility(LinuxFileSystem(role.host)) as fs:
            fs.write("/root/test", "content")
            with fs:
                fs.write("/root/test", "new_content")
                assert fs.read("/root/test") == "new_content"

            assert fs.read("/root/test") == "content"
    """
    # Mark utility as used so we do not call setup twice (one directly here, and
    # second time via @mh_utility_used decorator). Since the utility is created
    # on demand, we can always call the setup immediately.
    util._mh_utility_used = True
    mh_utility_setup(util)
    try:
        mh_utility_enter(util, "mh_utility")
        try:
            yield util
        finally:
            mh_utility_exit(util, "mh_utility")
    finally:
        mh_utility_teardown(util)


def mh_utility_setup_dependencies(
    obj: MultihostRole | MultihostHost,
    types: list[type[MultihostUtility]] = [MultihostUtility, MultihostReentrantUtility],
) -> None:
    """
    Setup all :class:`MultihostUtility` objects attributes of given object.

    :param obj: Multihost role or host object.
    :type obj: MultihostRole | MultihostHost
    :param types: Which MultihostUtility classes should be setup.
    :type types: list[type[MultihostUtility]]
    """
    for util in obj._mh_utility_dependencies:
        if not isinstance(util, tuple(types)):
            continue

        mh_utility_setup(util)
        mh_utility_enter(util, "mh_utility_dependencies")


def mh_utility_teardown_dependencies(
    obj: MultihostRole | MultihostHost,
    types: list[type[MultihostUtility]] = [MultihostUtility, MultihostReentrantUtility],
) -> None:
    """
    Teardown all :class:`MultihostUtility` objects attributes of given object.

    :param obj: Multihost role or host object.
    :type obj: MultihostRole | MultihostHost
    :param types: Which MultihostUtility classes should be setup.
    :type types: list[type[MultihostUtility]]
    """
    errors = []
    for util in reversed(obj._mh_utility_dependencies):
        if not isinstance(util, tuple(types)):
            continue

        try:
            mh_utility_exit(util, "mh_utility_dependencies")
        except Exception as e:
            errors.append(e)
        finally:
            try:
                mh_utility_teardown(util)
            except Exception as e:
                errors.append(e)

    if errors:
        raise TeardownExceptionGroup("Unable to teardown some utilities (util.teardown)", errors)


def mh_utility_enter_dependencies(obj: MultihostRole | MultihostHost, where: str) -> None:
    """
    Setup all :class:`MultihostUtility` objects attributes of given object.

    :param obj: Multihost role or host object.
    :type obj: MultihostRole | MultihostHost
    :param where: Where do we enter the utility.
    :type where: str
    """
    for util in obj._mh_utility_dependencies:
        util._op_state.clear(f"__enter__{where}")

    for util in obj._mh_utility_dependencies:
        util._op_state.set(f"__enter__{where}", "called")
        mh_utility_enter(util, where)


def mh_utility_exit_dependencies(obj: MultihostRole | MultihostHost, where: str) -> None:
    """
    Teardown all :class:`MultihostUtility` objects attributes of given object.

    :param obj: Multihost role or host object.
    :type obj: MultihostRole | MultihostHost
    :param where: Where do we enter the utility.
    :type where: str
    """
    errors = []
    for util in reversed(obj._mh_utility_dependencies):
        if not util._op_state.check(f"__enter__{where}", "called"):
            continue

        try:
            mh_utility_exit(util, where)
        except Exception as e:
            errors.append(e)

        util._op_state.clear(f"__enter__{where}")

    if errors:
        raise TeardownExceptionGroup("Unable to exit some utilities (util.__exit__)", errors)


def mh_utility_pytest_report_teststatus(
    obj: MultihostRole | MultihostHost, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
    """
    Run :meth:`MultihostUtility.pytest_report_teststatus` on all utilities.

    :param obj: Multihost role or host object.
    :type obj: MultihostRole | MultihostHost
    :param report: Pytest report.
    :type report: pytest.CollectReport | pytest.TestReport
    :param config: Pytest config.
    :type config: pytest.Config
    :return: Test status.
    :rtype: tuple[str, str, str | tuple[str, Mapping[str, bool]]] | None
    """
    for util in obj._mh_utility_dependencies:
        result = util.pytest_report_teststatus(report, config)
        if result is not None:
            return result

    return None



"""
.. avoid already-imported warning: PYTEST_DONT_REWRITE (hidden from sphinx)
"""

from __future__ import annotations

from ._private.artifacts import MultihostArtifactsType, MultihostHostArtifacts, MultihostTopologyControllerArtifacts
from ._private.data import MultihostItemData
from ._private.fixtures import (
    MultihostFixture,
    mh,
    mh_config,
    mh_logger,
    mh_topology,
    mh_topology_mark,
    mh_topology_name,
)
from ._private.logging import MultihostLogger
from ._private.marks import KnownTopologyBase, KnownTopologyGroupBase, TopologyMark
from ._private.multihost import (
    MultihostBackupHost,
    MultihostConfig,
    MultihostDomain,
    MultihostHost,
    MultihostOSFamily,
    MultihostReentrantUtility,
    MultihostRole,
    MultihostUtility,
    mh_utility,
    mh_utility_ignore_use,
    mh_utility_postpone_setup,
    mh_utility_used,
)
from ._private.plugin import MultihostPlugin, mh_fixture, pytest_addoption, pytest_configure
from ._private.topology import Topology, TopologyDomain
from ._private.topology_controller import BackupTopologyController, TopologyController

__all__ = [
    "mh",
    "mh_config",
    "mh_logger",
    "mh_topology",
    "mh_topology_mark",
    "mh_topology_name",
    "KnownTopologyBase",
    "KnownTopologyGroupBase",
    "MultihostLogger",
    "MultihostArtifactsType",
    "MultihostConfig",
    "MultihostDomain",
    "MultihostFixture",
    "MultihostHost",
    "MultihostBackupHost",
    "MultihostHostArtifacts",
    "MultihostItemData",
    "MultihostOSFamily",
    "MultihostPlugin",
    "MultihostRole",
    "MultihostTopologyControllerArtifacts",
    "MultihostUtility",
    "MultihostReentrantUtility",
    "mh_utility",
    "mh_utility_postpone_setup",
    "mh_utility_ignore_use",
    "mh_utility_used",
    "mh_fixture",
    "pytest_addoption",
    "pytest_configure",
    "Topology",
    "TopologyController",
    "BackupTopologyController",
    "TopologyDomain",
    "TopologyMark",
]



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Callable, Type, TypeAlias

from .conn import Powershell, Shell

__all__ = [
    "CLIBuilder",
]


class CLIBuilder(object):
    class option(Enum):
        """
        Command line parameter types.
        """

        PLAIN = auto()
        """
        Use plain parameter value without any modification.
        """

        VALUE = auto()
        """
        Use parameter value but enclose it in quotes in script mode.
        """

        SWITCH = auto()
        """
        Parameter is a switch which is enabled if value is True.
        """

        POSITIONAL = auto()
        """
        Parameter is a positional argument.
        """

    def __init__(self, shell: Shell) -> None:
        """
        :param shell: Target shell.
        :type shell: Shell
        """
        self.__shell: Shell = shell
        self.__prefix: str = "-" if self.__match_shell(Powershell) else "--"

    def command(self, command: str, args: CLIBuilderArgs) -> str:
        """
        Build full command line and return it as a string.

        Output can be passed directly to :meth:`Connection.run
        <pytest_mh.conn.Connection.run>` (``host.conn.run``).

        .. code-block:: python
            :caption: Example

            cli = CLIBuilder(Bash())
            args: CLIBuilderArgs = {
                "password": (cli.option.VALUE, None),     # None values are ignored
                "home": (cli.option.VALUE, "/home/jdoe"), # --home '/home/jdoe'
                "enabled": (cli.option.SWITCH, True),     # --enabled
                "login": (cli.option.POSITIONAL, "jdoe"), # 'jdoe'
            }

            line = cli.command("add-user", args)
            # add-user --home '/home/jdoe' --enabled 'jdoe'

        :param command: Command to call
        :type command: str
        :param args: Command's arguments
        :type args: CLIBuilderArgs
        :return: Full command line as string.
        :rtype: str
        """
        return " ".join(self.__build(command, args, quote_value=True))

    def argv(self, command: str, args: CLIBuilderArgs) -> list[str]:
        """
        Build full command line and return it it as list of arguments (full
        argv).

        Output can be passed directly to :meth:`Connection.exec
        <pytest_mh.conn.Connection.exec>` (``host.conn.exec``).

        .. code-block:: python
            :caption: Example

            cli = CLIBuilder(Bash())
            args: CLIBuilderArgs = {
                "password": (cli.option.VALUE, None),     # None values are ignored
                "home": (cli.option.VALUE, "/home/jdoe"), # --home '/home/jdoe'
                "enabled": (cli.option.SWITCH, True),     # --enabled
                "login": (cli.option.POSITIONAL, "jdoe"), # 'jdoe'
            }

            line = cli.argv("add-user", args)
            # ["add-user", "--home", "/home/jdoe", "--enabled", "jdoe"]

        :param command: Command to call
        :type command: str
        :param args: Command's arguments
        :type args: CLIBuilderArgs
        :return: Full command line as argv
        :rtype: list[str]
        """
        return self.__build(command, args, quote_value=False)

    def args(self, args: CLIBuilderArgs, *, quote_value=False) -> list[str]:
        """
        Build command's arguments and return them as a list (argv without
        command).

        Output can be used for additional processing by the caller.

        .. code-block:: python
            :caption: Example

            cli = CLIBuilder(Bash())
            args: CLIBuilderArgs = {
                "password": (cli.option.VALUE, None),     # None values are ignored
                "home": (cli.option.VALUE, "/home/jdoe"), # --home '/home/jdoe'
                "enabled": (cli.option.SWITCH, True),     # --enabled
                "login": (cli.option.POSITIONAL, "jdoe"), # 'jdoe'
            }

            line = cli.args(args)
            # ["--home", "/home/jdoe", "--enabled", "jdoe"]

            host.conn.run(f"user-add --encrypt-home {' '.join(line)}")

        :param args: Command's argument
        :type args: CLIBuilderArgs
        :param quote_value: True if values should enclosed with quotes, defaults to False
        :type quote_value: bool, optional
        :return: Arguments ready to use in command line (argv without command)
        :rtype: list[str]
        """
        return self.__build(None, args, quote_value)

    def __match_shell(self, shell: Type[Shell]):
        return isinstance(self.__shell, shell)

    def __build(self, command: str | None, args: CLIBuilderArgs, quote_value: bool) -> list[str]:
        def _get_option(name: str) -> str:
            return self.__prefix + name

        def _get_value(value: Any) -> str:
            return str(value) if not quote_value else f"'{value}'"

        def _add_argv(argv: list[str], key: str | None, value: Any, getvaluefn: Callable[[Any], str]) -> None:
            value = value if isinstance(value, list) else [value]
            for v in value:
                if key is not None:
                    argv.append(_get_option(key))

                argv.append(getvaluefn(v))

        argv = [command] if command is not None else []
        for key, item in args.items():
            if item is None:
                continue

            (type, value) = item
            if value is None:
                continue

            match type:
                case self.option.POSITIONAL:
                    _add_argv(argv, None, value, _get_value)
                case self.option.SWITCH:
                    if self.__match_shell(Powershell):
                        argv.append(f'{_get_option(key)}:{"$True" if value else "$False"}')
                    else:
                        if value:
                            argv.append(_get_option(key))
                case self.option.VALUE:
                    _add_argv(argv, key, value, _get_value)
                case self.option.PLAIN:
                    _add_argv(argv, key, value, str)
                case _:
                    raise ValueError(f"Unknown option type: {type}")

        return argv


CLIBuilderArgs: TypeAlias = dict[str, tuple[CLIBuilder.option, Any] | None]
"""CLIBuilder args format."""



from __future__ import annotations

import inspect
import logging
import sys
import textwrap
from functools import partial, wraps
from os import _exit
from pathlib import Path
from signal import SIGINT, signal
from typing import Callable, Generator, Literal, Type, get_type_hints

import pytest
import yaml

from .artifacts import MultihostArtifactsCollectable, MultihostArtifactsType
from .data import MultihostItemData
from .errors import TeardownExceptionGroup
from .fixtures import MultihostFixture
from .logging import MultihostLogger
from .marks import KnownTopologyBase, TopologyMark
from .multihost import (
    MultihostArtifactsMode,
    MultihostConfig,
    MultihostHost,
    MultihostReentrantUtility,
    MultihostRole,
    mh_utility_enter_dependencies,
    mh_utility_exit_dependencies,
    mh_utility_setup_dependencies,
    mh_utility_teardown_dependencies,
)
from .topology import Topology
from .topology_controller import TopologyController
from .types import MultihostOutcome

MarkStashKey = pytest.StashKey[TopologyMark | None]()


class MultihostPlugin(object):
    """
    Pytest multihost plugin.
    """

    def __init__(self, pytest_config: pytest.Config) -> None:
        self.config_class: Type[MultihostConfig] | None = None
        self.logger: logging.Logger = self._create_logger(pytest_config.option.verbose > 2)
        self.multihost: MultihostConfig | None = None
        self.topology: Topology | None = None
        self.confdict: dict | None = None
        self.current_mh: MultihostFixture | None = None
        self.current_topology: str | None = None
        self.required_hosts: list[MultihostHost] = []
        self.pytest_session: pytest.Session | None = None

        # CLI options
        self.mh_config: str = pytest_config.getoption("mh_config")
        self.mh_log_path: str = pytest_config.getoption("mh_log_path")
        self.mh_lazy_ssh: bool = pytest_config.getoption("mh_lazy_ssh")
        self.mh_topology: list[str] = pytest_config.getoption("mh_topology")
        self.mh_not_topology: list[str] = pytest_config.getoption("mh_not_topology")
        self.mh_exact_topology: bool = pytest_config.getoption("mh_exact_topology")
        self.mh_collect_artifacts: MultihostArtifactsMode = pytest_config.getoption("mh_collect_artifacts")
        self.mh_artifacts_dir: Path = Path(pytest_config.getoption("mh_artifacts_dir"))
        self.mh_compress_artifacts: bool = pytest_config.getoption("mh_compress_artifacts")
        self.mh_ignore_preferred_topology: bool = pytest_config.getoption("mh_ignore_preferred_topology")

        # Read --mh-collect-logs, default to --mh-collect-artifacts
        self.mh_collect_logs: MultihostArtifactsMode = pytest_config.getoption("mh_collect_logs")
        if self.mh_collect_logs is None:
            self.mh_collect_logs = self.mh_collect_artifacts

        # pytest options
        self.pytest_opt_collect_only: bool = pytest_config.getoption("collectonly")

    @classmethod
    def GetLogger(cls) -> logging.Logger:
        """
        Get plugin's logger.
        """

        return logging.getLogger("pytest_mh.plugin")

    def __load_conf(self, path: str) -> dict:
        """
        Load multihost configuration from a yaml file.

        :param path: Path to the yaml file.
        :type path: str
        :raises ValueError: If not file was provided.
        :raises IOError: If unable to read the file.
        :return: Parsed configuration.
        :rtype: dict

        :meta private:
        """
        if not path:
            raise ValueError("You need to provide valid multihost configuration file, use --mh-config=$path")

        try:
            with open(path, "r") as f:
                return yaml.safe_load(f)
        except Exception as e:
            raise IOError(f'Unable to open multihost configuration "{path}": {str(e)}')

    def sigint_handler(self, sig, frame) -> None:
        # This should not happen
        if self.multihost is None:
            self.logger.error("multihost can not be None, terminating")
            exit(1)

        # User really wants to exit now, ignore teardown
        if self.multihost._sigint:
            self.logger.info(
                "SIGINT received, terminating immediately. "
                "Teardown was not run, therefore hosts are in undefined state."
            )
            _exit(1)

        self.multihost._sigint = True

        # If we are in test, we can terminate gracefully immediately. Teardowns will run.
        if self.multihost._in_test:
            pytest.skip("SIGINT received, aborting running test.")

        # Otherwise we have to wait for setup/teardown to finish
        self.logger.info("")
        self.logger.info("")
        self.logger.info("SIGINT received SIGINT, I will finish current test and exit gracefully.")
        self.logger.info(
            "If you want to exit immediately, press CTRL-C one more time. "
            "In this case, however, hosts will end up in an undefined state "
            "since teardown methods will not run completely."
        )
        self.logger.info("")

    def setup(self) -> None:
        """
        Read and apply multihost configuration.

        :meta private:
        """
        if self.config_class is None:
            raise ValueError("Set MultihostPlugin.config_class to subclass of MultihostConfig")

        self.confdict = self.__load_conf(self.mh_config)

        logger = MultihostLogger.GetLogger()
        logger.setup(
            log_path=self.mh_log_path,
            artifacts_mode=self.mh_collect_logs,
            artifacts_dir=self.mh_artifacts_dir,
            confdict=self.confdict,
        )

        self.multihost = self.config_class(
            self.confdict,
            logger=logger,
            lazy_ssh=self.mh_lazy_ssh,
            artifacts_dir=self.mh_artifacts_dir,
            artifacts_mode=self.mh_collect_artifacts,
            artifacts_compression=self.mh_compress_artifacts,
        )
        self.topology = Topology.FromMultihostConfig(self.confdict)

    @pytest.hookimpl(trylast=True)
    def pytest_sessionstart(self, session: pytest.Session) -> None:
        """
        Setup the module and log information about given multihost configuration
        and provided options.

        :meta private:
        """
        self.pytest_session = session

        # Calling the setup here instead of in constructor to allow running
        # pytest --help and other action-less parameters.
        self.setup()

        # Silent mypy false positive
        if self.topology is None:
            raise ValueError("Topology must be already set!")

        if self.multihost is None:
            self.logger.info(self._fmt_bold("Multihost configuration:"))
            self.logger.info("  No multihost configuration provided.")
            self.logger.info("  Make sure to run tests with --mh-log-path parameter.")
            self.logger.info("")
            return

        self.logger.info(self._fmt_bold("Multihost configuration:"))
        self.logger.info(textwrap.indent(yaml.dump(self.confdict, sort_keys=False), "  "))
        self.logger.info(self._fmt_bold("Detected topology:"))
        self.logger.info(textwrap.indent(yaml.dump(self.topology.export(), sort_keys=False), "  "))
        self.logger.info(self._fmt_bold("Additional settings:"))
        self.logger.info(f"  config file: {self.mh_config}")
        self.logger.info(f"  log path: {self.mh_log_path}")
        self.logger.info(f"  lazy ssh: {self.mh_lazy_ssh}")
        self.logger.info(f"  topology filter: {', '.join(self.mh_topology + [f'!{x}' for x in self.mh_not_topology])}")
        self.logger.info(f"  require exact topology: {self.mh_exact_topology}")
        self.logger.info(f"  collect artifacts: {self.mh_collect_artifacts}")
        self.logger.info(f"  artifacts directory: {self.mh_artifacts_dir}")
        self.logger.info(f"  collect logs: {self.mh_collect_logs}")
        self.logger.info(f"  ignore-preferred-topology: {self.mh_ignore_preferred_topology}")
        self.logger.info("")

        signal(SIGINT, self.sigint_handler)

    @pytest.hookimpl(trylast=True)
    def pytest_sessionfinish(self, session: pytest.Session, exitstatus: int | pytest.ExitCode) -> None:
        """
        Teardown the module.

        :meta private:
        """
        if self.multihost is None:
            return

        # Run pytest_teardown on all hosts required by selected tests
        if not self.pytest_opt_collect_only:
            self._teardown_hosts(self.required_hosts)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_make_collect_report(self, collector: pytest.Collector) -> Generator[None, pytest.CollectReport, None]:
        """
        If multiple topology marks are present on the collected test, we need to
        parametrize it. In order to do so, the test is replaced with multiple
        clones, one for each topology.

        The topology associated with the test clone is stored in
        ``topology_mark`` property of the clone.

        :meta private:
        """

        outcome = yield
        report = outcome.get_result()

        if not report.result:
            return

        if self.multihost is None:
            return

        new_result = []
        for result in report.result:
            if not isinstance(result, pytest.Function):
                new_result.append(result)
                continue

            has_marks = False
            for mark in self.multihost.TopologyMarkClass.ExpandMarkers(result):
                has_marks = True
                topology_mark = self.multihost.TopologyMarkClass.Create(result, mark)
                f = self._clone_function(f"{result.name} ({topology_mark.name})", result)
                f.stash[MarkStashKey] = topology_mark
                new_result.append(f)

            if not has_marks:
                result.stash[MarkStashKey] = None
                new_result.append(result)

        report.result = new_result

    @pytest.hookimpl(hookwrapper=True)
    def pytest_collection_modifyitems(self, config: pytest.Config, items: list[pytest.Item]) -> Generator:
        """
        Filter collected items and deselect these that can not be run on the
        selected multihost configuration.

        Internal plugin's data are stored in ``multihost`` property of each
        :class:`pytest.Item`.

        :meta private:
        """
        data: MultihostItemData | None = None
        selected: list[pytest.Item] = []
        deselected: list[pytest.Item] = []
        mapping: dict[str, list[pytest.Item]] = {}

        # Silent mypy false positive
        if self.multihost is None:
            return

        for item in items:
            data = MultihostItemData(self.multihost, item.stash[MarkStashKey]) if self.multihost else None
            MultihostItemData.SetData(item, data)

            if not self._can_run_test(item, data):
                deselected.append(item)
                continue

            # This test can be run, perform delayed initialization of data.
            if data is not None:
                data._init()

            # Map test items by topology name so we can sort them later
            if data is None or data.topology_mark is None:
                mapping.setdefault("", []).append(item)
            else:
                mapping.setdefault(data.topology_mark.name, []).append(item)

        # Sort test by topology name
        selected = sum([y for _, y in sorted(mapping.items())], [])

        # Yield result to pytest
        config.hook.pytest_deselected(items=deselected)
        items[:] = selected

        yield

        # List of items may have been further modified by other plugins or filters
        # Remember all hosts required to run selected tests
        required_hosts_set: set[MultihostHost] = set()
        for item in items:
            data = MultihostItemData.GetData(item)
            if data is None or data.topology_mark is None:
                continue

            required_hosts_set.update(self.multihost.topology_hosts(data.topology_mark.topology))

        # Sort required host by name to provide deterministic runs
        self.required_hosts = sorted(required_hosts_set, key=lambda x: x.hostname)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_collection_finish(self, session: pytest.Session) -> Generator:
        # Log required hosts
        self.logger.info("")
        self.logger.info("")
        self.logger.info(self._fmt_bold("Selected tests will use the following hosts:"))
        for host in self.required_hosts:
            self.logger.info(f"  {host.role}: {host.hostname}")
        self.logger.info("")

        yield

        # Run pytest_setup on all hosts required by selected tests
        if not self.pytest_opt_collect_only:
            # Connect to all required hosts to fail quickly if some connection
            # cannot be established.
            if self.multihost is not None and not self.multihost.lazy_ssh:
                for host in self.required_hosts:
                    host.conn.connect()

            self._setup_hosts(self.required_hosts)

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_setup(self, item: pytest.Item) -> None:
        """
        Create fixtures requested in :class:`~pytest_mh.TopologyMark`
        (``@pytest.mark.topology``). It adds the fixture names into ``funcargs``
        property of the pytest item in order to make them available.

        At this step, the fixtures do not have any value. The value is assigned
        later in :func:`pytest_runtest_call` hook.

        :meta private:
        """
        if not isinstance(item, pytest.Function):
            return

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        if self.multihost is None or data is None or data.topology_mark is None:
            return
        mark: TopologyMark = data.topology_mark

        if self.multihost._sigint:
            pytest.exit("Aborted because SIGINT was received.")

        # Execute per-topology setup if topology is switched.
        if self._topology_switch(None, item):
            self.current_topology = mark.name
            self._setup_topology(mark.name, mark.controller)

        if not mark.controller._op_state.check_success("topology_setup"):
            pytest.skip("Error in topology setup")

        # Make mh fixture always available
        if "mh" not in item.fixturenames:
            item.fixturenames.append("mh")

        # Fill in parameters that will be set later in pytest_runtest_call hook,
        # otherwise pytest will raise unknown fixture error.
        spec = inspect.getfullargspec(item.obj)
        for arg in mark.args:
            if arg in spec.args:
                item.funcargs[arg] = None

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_call(self, item: pytest.Item) -> None:
        """
        Assign values to dynamically created multihost fixtures.

        :meta private:
        """
        if not isinstance(item, pytest.Function):
            return

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        if data is None:
            return

        if data.topology_mark is not None:
            mh = item.funcargs["mh"]
            if not isinstance(mh, MultihostFixture):
                raise ValueError(f"Fixture mh is not MultihostFixture but {type(mh)}!")

            self.current_mh = mh
            data.topology_mark.apply(mh, item.funcargs)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_teardown(self, item: pytest.Item, nextitem: pytest.Item | None) -> Generator:
        """
        Teardown topology if we detect a topology switch.

        :meta private:
        """
        # Inner teardown callback may raise error, but we still need to call
        # topology teardown if needed.
        try:
            yield
        finally:
            self.current_mh = None

            if self.multihost is None:
                return

            data: MultihostItemData | None = MultihostItemData.GetData(item)
            if data is None or data.topology_mark is None:
                return
            mark: TopologyMark = data.topology_mark

            # Execute per-topology teardown if topology changed.
            if self._topology_switch(item, nextitem) or self.multihost._sigint:
                self._teardown_topology(mark.name, mark.controller)

    @pytest.hookimpl(tryfirst=True)
    def pytest_report_teststatus(
        self, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
    ) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
        if report.when != "call":
            return None

        if hasattr(report, "_pytest_mh__teststatus"):
            return report._pytest_mh__teststatus

        if self.current_mh is None:
            return None

        # Store current outcome in case it is changed by the hook
        original_outcome = report.outcome

        status = self.current_mh._pytest_report_teststatus(report, config)
        setattr(report, "_pytest_mh__teststatus", status)

        # If the outcome is changed and failed, count it towards failures.
        if original_outcome != report.outcome and report.failed:
            if self.pytest_session is not None:
                self.pytest_session.testsfailed += 1

        return status

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(
        self, item: pytest.Item, call: pytest.CallInfo[None]
    ) -> Generator[None, pytest.TestReport, None]:
        """
        Store test outcome in multihost data: item.multihost.outcome. The outcome
        can be 'passed', 'failed' or 'skipped'.
        """
        outcome = yield

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        result: pytest.TestReport = outcome.get_result()

        if data is None:
            raise RuntimeError("MultihostItemData should not be None")

        if result.when != "call":
            return

        data.outcome = result.outcome
        data.result = result

    # Hook from pytest-output plugin
    @pytest.hookimpl(optionalhook=True)
    def pytest_output_item_collected(self, config: pytest.Config, item) -> None:
        try:
            from pytest_output.output import OutputDataItem
        except ImportError:
            pass

        if not isinstance(item, OutputDataItem):
            raise ValueError(f"Unexpected item type: {type(item)}")

        data: MultihostItemData | None = MultihostItemData.GetData(item.item)
        if data is None or data.topology_mark is None:
            return

        item.extra.setdefault("pytest-mh", {})["Topology"] = data.topology_mark.name

    def _fmt_color(self, text: str, color: str) -> str:
        if sys.stdout.isatty():
            reset = "\033[0m"
            return f"{color}{text}{reset}"

        return text

    def _fmt_bold(self, text: str) -> str:
        return self._fmt_color(text, "\033[1m")

    def _create_logger(self, verbose) -> logging.Logger:
        stdout = logging.StreamHandler(sys.stdout)
        stdout.setLevel(logging.DEBUG)
        stdout.setFormatter(logging.Formatter("%(message)s"))

        logger = self.GetLogger()
        logger.addHandler(stdout)
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)

        return logger

    def _is_multihost_required(self, item: pytest.Item) -> bool:
        return item.get_closest_marker(name="topology") is not None

    def _can_run_preferred_topology(self, mark: pytest.Mark, current_topology: str, item: pytest.Item) -> bool:
        if len(mark.args) != 1:
            raise ValueError(
                f"{item.nodeid}: Unexpected number of arguments to pytest.mark.preferred_topology: "
                f"got {len(mark.args)}, expected 1"
            )

        arg = mark.args[0]

        if isinstance(arg, KnownTopologyBase):
            name = arg.value.name
        elif isinstance(arg, TopologyMark):
            name = arg.name
        elif isinstance(arg, str):
            name = arg
        else:
            raise ValueError(
                f"{item.nodeid}: Unexpected type of pytest.mark.preferred_topology: "
                f"got {type(arg)}, expected KnownTopologyBase | TopologyMark | str"
            )

        if name == current_topology:
            return True

        return False

    def _can_run_test(self, item: pytest.Item, data: MultihostItemData | None) -> bool:
        if data is None:
            return not self._is_multihost_required(item)

        if self.topology is None:
            raise ValueError("Topology must be already set!")

        if data.topology_mark is not None:
            if self.mh_exact_topology:
                if data.topology_mark.topology != self.topology:
                    return False
            else:
                if not self.topology.satisfies(data.topology_mark.topology):
                    return False

        if self.mh_not_topology:
            if data.topology_mark is not None and data.topology_mark.name in self.mh_not_topology:
                return False

        if self.mh_topology:
            if data.topology_mark is None:
                return False

            if data.topology_mark.name not in self.mh_topology:
                return False

        # Run only for preferred topology unless specific topology is requested or the marker is ignored
        if not self.mh_topology and not self.mh_ignore_preferred_topology:
            preferred_topology = item.get_closest_marker(name="preferred_topology")
            if preferred_topology is not None and data.topology_mark is not None:
                return self._can_run_preferred_topology(preferred_topology, data.topology_mark.name, item)

        return True

    def _clone_function(self, name: str, f: pytest.Function) -> pytest.Function:
        callspec = f.callspec if hasattr(f, "callspec") else None

        return pytest.Function.from_parent(
            parent=f.parent,
            name=name,
            callspec=callspec,
            callobj=f.obj,
            keywords=f.keywords,
            fixtureinfo=f._fixtureinfo,
            originalname=f.originalname,
        )

    def _topology_switch(self, curitem: pytest.Item | None, nextitem: pytest.Item | None) -> bool:
        # No more items means topology switch for our usecase
        if nextitem is None:
            return True

        # If current item is None, we need to check current topology
        if curitem is None:
            # This is a first test in the new topology
            if self.current_topology is None:
                return True
            # We always set current_topology to None when switching topologies
            else:
                return False

        curdata: MultihostItemData | None = MultihostItemData.GetData(curitem)
        nextdata: MultihostItemData | None = MultihostItemData.GetData(nextitem)

        if curdata is None or nextdata is None:
            raise RuntimeError("Data can not be None")

        # If the test does not have topology marker, we consider it a switch
        if curdata.topology_mark is None or nextdata.topology_mark is None:
            return True

        # Different topology name is a switch
        if curdata.topology_mark.name != nextdata.topology_mark.name:
            return True

        return False

    def _setup_hosts(self, hosts: list[MultihostHost]) -> None:
        # Silent mypy false positive
        if self.multihost is None:
            raise RuntimeError("Multihost configuration is not present.")

        for host in hosts:
            outcome: MultihostOutcome = "error"
            try:
                try:
                    host.logger.phase(f"PYTEST SETUP HOST UTILS :: {host.hostname}")
                    mh_utility_setup_dependencies(host, [MultihostReentrantUtility])
                    host._op_state.set_success("pytest_setup_utils")
                finally:
                    host.logger.phase(f"PYTEST SETUP HOST UTILS DONE :: {host.hostname}")

                try:
                    host.logger.phase(f"PYTEST SETUP :: {host.hostname}")
                    host.pytest_setup()
                    host._op_state.set_success("pytest_setup")
                    outcome = "passed"
                finally:
                    host.logger.phase(f"PYTEST SETUP DONE :: {host.hostname}")
            finally:
                self._collect_artifacts(
                    id=host.hostname,
                    hostdir=False,
                    type="pytest_setup",
                    path=f"hosts/{host.hostname}/pytest_setup",
                    collectable={host: [host, *host._mh_utility_dependencies]},
                    outcome=outcome,
                    logger=host.logger,
                )

                self.multihost.logger.flush(outcome, f"hosts/{host.hostname}/pytest_setup.log")

    def _teardown_hosts(self, hosts: list[MultihostHost]) -> None:
        # Silent mypy false positive
        if self.multihost is None:
            raise RuntimeError("Multihost configuration is not present.")

        errors = []
        for host in hosts:
            outcome: MultihostOutcome = "error"
            try:
                try:
                    host.logger.phase(f"PYTEST TEARDOWN :: {host.hostname}")
                    if host._op_state.check_success("pytest_setup"):
                        host.pytest_teardown()
                except Exception as e:
                    errors.append(e)
                finally:
                    host.logger.phase(f"PYTEST TEARDOWN DONE :: {host.hostname}")

                try:
                    host.logger.phase(f"PYTEST TEARDOWN HOST UTILS :: {host.hostname}")
                    mh_utility_teardown_dependencies(host, [MultihostReentrantUtility])
                    outcome = "passed"
                except Exception as e:
                    errors.append(e)
                finally:
                    host.logger.phase(f"PYTEST TEARDOWN HOST UTILS DONE :: {host.hostname}")
            finally:
                self._collect_artifacts(
                    id=host.hostname,
                    hostdir=False,
                    type="pytest_teardown",
                    path=f"hosts/{host.hostname}/pytest_teardown",
                    collectable={host: [host, *host._mh_utility_dependencies]},
                    outcome=outcome,
                    logger=host.logger,
                )

                self.multihost.logger.flush(outcome, f"hosts/{host.hostname}/pytest_teardown.log")

        if errors:
            raise TeardownExceptionGroup("Unable to teardown some hosts (host.pytest_teardown)", errors)

    def _setup_topology(self, name: str, controller: TopologyController) -> None:
        # Silent mypy false positive
        if self.multihost is None:
            raise RuntimeError("Multihost configuration is not present.")

        outcome: MultihostOutcome = "error"
        try:
            try:
                controller.logger.phase(f"TOPOLOGY SETUP ENTER HOST UTILS :: {name}")
                for host in controller.hosts:
                    mh_utility_enter_dependencies(host, "topology_setup")
            finally:
                controller.logger.phase(f"TOPOLOGY SETUP ENTER HOST UTILS DONE :: {name}")

            try:
                controller.logger.phase(f"TOPOLOGY SETUP :: {name}")
                controller._invoke_with_args(controller.set_artifacts)
                controller._invoke_with_args(controller.topology_setup)
                controller._op_state.set_success("topology_setup")
                outcome = "passed"
            finally:
                controller.logger.phase(f"TOPOLOGY SETUP DONE :: {name}")
        finally:
            self._collect_artifacts(
                id=name,
                hostdir=True,
                type="topology_setup",
                path=f"topologies/{name}/topology_setup",
                collectable={x: [x, *x._mh_utility_dependencies, controller] for x in controller.hosts},
                outcome=outcome,
                logger=controller.logger,
            )

            self.multihost.logger.flush(outcome, f"topologies/{name}/topology_setup.log")

    def _teardown_topology(self, name: str, controller: TopologyController) -> None:
        # Silent mypy false positive
        if self.multihost is None:
            raise RuntimeError("Multihost configuration is not present.")

        outcome: MultihostOutcome = "error"
        try:
            errors = []
            try:
                controller.logger.phase(f"TOPOLOGY TEARDOWN :: {name}")
                if controller._op_state.check_success("topology_setup"):
                    controller._invoke_with_args(controller.topology_teardown)
            except Exception as e:
                errors.append(e)
            finally:
                self.current_topology = None
                controller.logger.phase(f"TOPOLOGY TEARDOWN DONE :: {name}")

            controller.logger.phase(f"TOPOLOGY TEARDOWN EXIT HOST UTILS :: {name}")
            for host in controller.hosts:
                try:
                    mh_utility_exit_dependencies(host, "topology_setup")
                except Exception as e:
                    errors.append(e)
            controller.logger.phase(f"TOPOLOGY TEARDOWN EXIT HOST UTILS DONE :: {name}")

            if errors:
                raise TeardownExceptionGroup(
                    "Unable to teardown topology (topology_controller.topology_teardown)", errors
                )

            outcome = "passed"
        finally:
            self._collect_artifacts(
                id=name,
                hostdir=True,
                type="topology_teardown",
                path=f"topologies/{name}/topology_teardown",
                collectable={x: [x, *x._mh_utility_dependencies, controller] for x in controller.hosts},
                outcome=outcome,
                logger=controller.logger,
            )

            self.multihost.logger.flush(outcome, f"topologies/{name}/topology_teardown.log")

    def _collect_artifacts(
        self,
        *,
        id: str,
        hostdir: bool,
        type: MultihostArtifactsType,
        path: str,
        collectable: dict[MultihostHost, list[MultihostArtifactsCollectable]],
        outcome: MultihostOutcome,
        logger: MultihostLogger,
    ) -> None:
        logger.phase(f"COLLECT ARTIFACTS :: {id}")
        for host, objects in collectable.items():
            dest = f"{path}/{host.hostname}" if hostdir else path
            try:
                host.artifacts_collector.collect(type, path=dest, outcome=outcome, collect_objects=objects)
            except Exception as e:
                self.logger.error(
                    "An error happend when collecting artifacts",
                    extra={
                        "data": {
                            "Error message": str(e),
                        }
                    },
                )
        logger.phase(f"COLLECT ARTIFACTS DONE :: {id}")


# These pytest hooks must be available outside of the plugin's class because
# they are executed before the plugin is registered.


def pytest_addoption(parser):
    """
    Pytest hook: add command line options.
    """

    parser.addoption("--mh-config", action="store", help="Path to the multihost configuration file")

    parser.addoption("--mh-log-path", action="store", help="Path to store multihost logs")

    parser.addoption("--mh-lazy-ssh", action="store_true", help="Postpone connecting to host SSH until it is required")

    parser.addoption(
        "--mh-ignore-preferred-topology",
        action="store_true",
        help="All topologies will run, ignore the preferred_topology marker",
    )

    parser.addoption(
        "--mh-topology",
        action="append",
        default=[],
        help="Filter tests by given topology, can be set multiple times",
    )

    parser.addoption(
        "--mh-not-topology",
        action="append",
        default=[],
        help="Do not run tests for given topology, can be set multiple times",
    )

    parser.addoption(
        "--mh-exact-topology",
        action="store_true",
        help="Test will be deselected if its topology does not match given multihost config exactly",
    )

    parser.addoption(
        "--mh-collect-artifacts",
        action="store",
        default="on-failure",
        nargs="?",
        choices=["never", "on-failure", "always"],
        help="Collect artifacts after test run (default: %(default)s)",
    )

    parser.addoption(
        "--mh-artifacts-dir",
        action="store",
        default="./artifacts",
        help="Directory where artifacts will be stored (default: %(default)s)",
    )

    parser.addoption(
        "--mh-compress-artifacts",
        action="store_true",
        help="If set, test artifacts are stored in a compressed archive",
    )

    parser.addoption(
        "--mh-collect-logs",
        action="store",
        default=None,
        nargs="?",
        choices=["never", "on-failure", "always"],
        help="Collect logs mode (default: use value of --mh-collect-artifacts)",
    )


def pytest_configure(config: pytest.Config):
    """
    Pytest hook: register multihost plugin.
    """

    # register additional markers
    config.addinivalue_line(
        "markers",
        "topology(name: str, topology: pytest_mh.Topology, domains: dict[str, str], /, "
        + "*, fixture1=target1, ...): topology required to run the test",
    )

    config.addinivalue_line(
        "markers",
        "require(condition, reason): evaluate condition, parameters may be topology fixture, "
        "the test is skipped if condition is not met",
    )

    config.addinivalue_line(
        "markers",
        "preferred_topology(topology: KnownTopologyBase | TopologyMark | str): "
        "mark test with a preferred topology."
        "Test will execute once, skipping additional topologies",
    )

    config.pluginmanager.register(MultihostPlugin(config), "MultihostPlugin")


def mh_fixture(fixture_function: Callable | None = None, *, scope: Literal["function"] = "function"):
    """
    This creates a function-scoped pytest fixture that can access MultihostRole
    objects that are available to the test directly.

    .. note::

        For this to work correctly, all multihost fixtures have to be correctly
        typed. It will not work without the type hints.

    At this moment, only ``function`` scope is supported.

    .. code-block:: python

        @mh_fixture()
        def my_fixture(client: Client, request: pytest.FixtureRequest):
            pass

        @pytest.mark.topology(KnownTopology.LDAP)
        def test_example(client: Client, ldap: LDAP, my_fixture):
            pass

    :param scope: Fixture scope, defaults to "function"
    :type scope: Literal[&quot;function&quot;], optional
    """

    def decorator(fn):
        full_sig = inspect.signature(fn)
        mh_args = []
        for arg, hint in get_type_hints(fn).items():
            if issubclass(hint, MultihostRole):
                mh_args.append(arg)
                continue

        def call_fixture(mh: MultihostFixture, *args, **kwargs):
            if "mh" in full_sig.parameters:
                kwargs["mh"] = mh

            for arg in mh_args:
                if arg not in mh.fixtures:
                    raise KeyError(f"{fn.__name__}: Parameter {arg} is not a valid topology fixture")

                kwargs[arg] = mh.fixtures[arg]

            return fn(*args, **kwargs)

        @wraps(fn)
        def wrapper_normal(mh: MultihostFixture, *args, **kwargs):
            return call_fixture(mh, *args, **kwargs)

        @wraps(fn)
        def wrapper_yield(mh: MultihostFixture, *args, **kwargs):
            gen = call_fixture(mh, *args, **kwargs)
            yield next(gen)
            try:
                yield next(gen)
            except StopIteration:
                pass

        # Select wrapper
        wrapper = wrapper_normal
        if inspect.isgeneratorfunction(fn):
            wrapper = wrapper_yield

        # Bound multihost parameters
        cb = wraps(fn)(partial(wrapper, **{arg: None for arg in mh_args}))

        # Create pytest fixture
        fixture = pytest.fixture(scope="function")(cb)

        # Mock parameters so they are correctly recognized by pytest fixture
        partial_parameters = [inspect.Parameter("mh", inspect._POSITIONAL_OR_KEYWORD)]
        partial_parameters.extend(
            [param for key, param in full_sig.parameters.items() if key != "mh" and key not in mh_args]
        )

        # pytest < 8.4.0 provides its own wrapper around the function and checks
        # signature of the real function, not the partial one, so we need to mock
        # signatures of the decorated function (fn)
        if hasattr(fixture, "__pytest_wrapped__"):
            func = fixture.__pytest_wrapped__.obj.func
        # pytest >= 8.4.0 does not unwrap it anymore and reads signature of the
        # provided function, i.e. the partial one (cb)
        elif hasattr(fixture, "__wrapped__"):
            func = fixture._get_wrapped_function()
        else:
            raise AttributeError(
                "Fixture object has no __pytest_wrapped__ nor __wrapped__ attribute, "
                "report this to pytest-mh upstream."
            )

        func.__signature__ = inspect.Signature(partial_parameters, return_annotation=full_sig.return_annotation)

        return fixture

    # Direct decoration.
    if fixture_function:
        return decorator(fixture_function)

    return decorator



from __future__ import annotations

from functools import partial, wraps
from types import SimpleNamespace
from typing import Any, Callable, Generic

from .artifacts import MultihostArtifactsType, MultihostTopologyControllerArtifacts
from .logging import MultihostLogger
from .misc import OperationStatus, invoke_callback
from .multihost import ConfigType, MultihostBackupHost, MultihostDomain, MultihostHost
from .topology import Topology, TopologyDomain


class TopologyController(Generic[ConfigType]):
    """
    Topology controller can be associated with a topology via TopologyMark to
    provide additional per-topology hooks such as per-topology setup and
    teardown.

    When inheriting from this class, keep it mind that there is postponed
    initialization of all present properties therefore you can not access them
    inside the constructor. The properties are initialized when a test is
    collected. Override :meth:`init` instead of the constructor if you need to
    access these properties from constructor.

    Each method can take MultihostHost object as parameters as defined in
    topology fixtures.

    .. code-block:: python
        :caption: Example topology controller

        class ExampleController(TopologyController):
            def set_artifacts(self, client: ClientHost) -> None:
                self.artifacts.topology_setup[client] = {"/etc/issue"}

            def skip(self, client: ClientHost) -> str | None:
                result = client.conn.run(
                    ''' # Implement your requirement check here exit 1 ''',
                    raise_on_error=False)
                if result.rc != 0:
                    return "Topology requirements were not met"

                return None

            def topology_setup(self, client: ClientHost):
                # One-time setup, prepare the host for this topology # Changes
                done here are shared for all tests pass

            def topology_teardown(self, client: ClientHost):
                # One-time teardown, this should undo changes from #
                topology_setup pass

            def setup(self, client: ClientHost):
                # Perform per-topology test setup # This is called before
                execution of every test pass

            def teardown(self, client: ClientHost):
                # Perform per-topology test teardown, this should undo changes #
                from setup pass

    .. code-block:: python
        :caption: Example with low-level topology mark

        class ExampleController(TopologyController):
            # Implement methods you are interested in here pass

        @pytest.mark.topology(
            "example", Topology(TopologyDomain("example", client=1)),
            controller=ExampleController(),
            fixtures=dict(client="example.client[0]")
        ) def test_example(client: Client):
            pass

    .. code-block:: python
        :caption: Example with KnownTopology

        class ExampleController(TopologyController):
            # Implement methods you are interested in here pass

        @final @unique class KnownTopology(KnownTopologyBase):
            EXAMPLE = TopologyMark(
                name='example', topology=Topology(TopologyDomain("example",
                client=1)), controller=ExampleController(),
                fixtures=dict(client='example.client[0]'),
            )

        @pytest.mark.topology(KnownTopology.EXAMPLE) def test_example(client:
        Client):
            pass
    """

    def __init__(self) -> None:
        self._op_state: OperationStatus = OperationStatus()
        """Keep state of setup and teardown methods."""

        self.__name: str | None = None
        self.__multihost: ConfigType | None = None
        self.__logger: MultihostLogger | None = None
        self.__topology: Topology | None = None
        self.__ns: SimpleNamespace | None = None
        self.__args: dict[str, MultihostHost | list[MultihostHost]] | None = None
        self.__initialized: bool = False

        self.artifacts: MultihostTopologyControllerArtifacts = MultihostTopologyControllerArtifacts()
        """
        List of artifacts that will be automatically collected at specific
        places. This list can be dynamically extended. Values may contain
        wildcard character.
        """

    def init(
        self,
        name: str,
        multihost: ConfigType,
        logger: MultihostLogger,
        topology: Topology,
        mapping: dict[str, str],
    ):
        """
        Postponed initialization of the topology controller, called by the
        plugin.

        All properties are set and accessible after this method is finished.

        :param name: Topology name.
        :type name: str
        :param multihost: MultihostConfig instance.
        :type multihost: ConfigType
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param topology: Topology.
        :type topology: Topology
        :param mapping: Host to fixtures mapping.
        :type mapping: dict[str, str]
        """
        # This is called for each testcase but the controller may be shared with
        # multiple testcases therefore we want to avoid multiple initialization.
        if self.__initialized:
            return

        self.__name = name
        self.__multihost = multihost
        self.__logger = logger
        self.__topology = topology
        self.__ns, self.__args, self.__hosts = self._build_namespace_and_args(multihost.domains, topology, mapping)

        self.__initialized = True

    def _build_namespace_and_args(
        self,
        mh_domains: list[MultihostDomain],
        topology: Topology,
        mapping: dict[str, str],
    ) -> tuple[SimpleNamespace, dict[str, MultihostHost | list[MultihostHost]], list[MultihostHost]]:
        root = SimpleNamespace()
        paths: dict[str, MultihostHost | list[MultihostHost]] = {}
        hosts: set[MultihostHost] = set()

        for mh_domain in mh_domains:
            if mh_domain.id in topology:
                ns, nspaths, nshosts = self._build_domain_namespace_and_paths(topology.get(mh_domain.id), mh_domain)
                setattr(root, mh_domain.id, ns)
                paths.update(**nspaths)
                hosts.update(nshosts)

        args = {}
        for name, path in mapping.items():
            args[name] = paths[path]

        return (root, args, sorted(hosts, key=lambda x: x.hostname))

    def _build_domain_namespace_and_paths(
        self,
        topology_domain: TopologyDomain,
        mh_domain: MultihostDomain,
    ) -> tuple[SimpleNamespace, dict[str, MultihostHost | list[MultihostHost]], set[MultihostHost]]:
        ns = SimpleNamespace()
        paths: dict[str, MultihostHost | list[MultihostHost]] = {}
        domain_hosts: set[MultihostHost] = set()

        for role_name in mh_domain.roles:
            if role_name not in topology_domain:
                continue

            count = topology_domain.get(role_name)
            hosts = [host for host in mh_domain.hosts_by_role(role_name)[:count]]
            domain_hosts.update(hosts)
            setattr(ns, role_name, hosts)

            paths[f"{topology_domain.id}.{role_name}"] = hosts
            for index, host in enumerate(hosts):
                paths[f"{topology_domain.id}.{role_name}[{index}]"] = host

        return (ns, paths, domain_hosts)

    def _invoke_with_args(self, cb: Callable) -> Any:
        if self.__args is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return invoke_callback(cb, **self.__args)

    @property
    def name(self) -> str:
        """
        Topology name.

        This property cannot be accessed from the constructor.

        :return: Topology name.
        :rtype: str
        """
        if self.__name is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__name

    @property
    def topology(self) -> Topology:
        """
        Multihost topology.

        This property cannot be accessed from the constructor.

        :return: Topology.
        :rtype: Topology
        """
        if self.__topology is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__topology

    @property
    def multihost(self) -> ConfigType:
        """
        Multihost configuration.

        This property cannot be accessed from the constructor.

        :return: Multihost configuration.
        :rtype: MultihostConfig
        """
        if self.__multihost is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__multihost

    @property
    def logger(self) -> MultihostLogger:
        """
        Multihost logger.

        This property cannot be accessed from the constructor.

        :return: Multihost logger.
        :rtype: MultihostLogger
        """
        if self.__logger is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__logger

    @property
    def ns(self) -> SimpleNamespace:
        """
        Namespace of MultihostHost objects accessible by domain id and roles names.

        This property cannot be accessed from the constructor.

        :return: Namespace.
        :rtype: SimpleNamespace
        """
        if self.__ns is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__ns

    @property
    def hosts(self) -> list[MultihostHost]:
        """
        List of MultihostHost objects available in this topology.

        This property cannot be accessed from the constructor.

        :return: List of MultihostHost objects.
        :rtype: list[MultihostHost]
        """
        if self.__hosts is None:
            raise RuntimeError("TopologyController has not been initialized yet")

        return self.__hosts

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Return the list of artifacts to collect.

        This just returns :attr:`artifacts`, but it is possible to override this
        method in order to generate additional artifacts that were not created
        by the test, or detect which artifacts were created and update the
        artifacts list.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        return self.artifacts.get(host, artifacts_type)

    def set_artifacts(self, *args, **kwargs) -> None:
        """
        Called before :meth:`topology_setup` to set topology artifacts.

        Note that the artifacts can be set in any other method as well. This
        dedicated method is just for your convenience.
        """
        return

    def skip(self, *args, **kwargs) -> str | None:
        """
        Called before a test is executed.

        If a non-None value is returned the test is skipped, using the returned
        value as a skip reason.

        :rtype: str | None
        """
        return None

    def topology_setup(self, *args, **kwargs) -> None:
        """
        Called once before executing the first test of given topology.
        """
        pass

    def topology_teardown(self, *args, **kwargs) -> None:
        """
        Called once after all tests for given topology were run.
        """
        pass

    def setup(self, *args, **kwargs) -> None:
        """
        Called before execution of each test.
        """
        pass

    def teardown(self, *args, **kwargs) -> None:
        """
        Called after execution of each test.
        """
        pass


class BackupTopologyController(TopologyController[ConfigType]):
    """
    Implements automatic backup and restore of all topology hosts that inherit
    from :class:`MultihostBackupHost`.

    The backup of all hosts is taken in :meth:`topology_setup`. It is expected
    that this method is overridden by the user to setup the topology
    environment. In such case, it is possible to call
    ``super().topology_setup(**kwargs)`` at the end of the overridden function
    or omit this call and store the backup in :attr:`backup_data` manually.

    :meth:`teardown` restores the hosts to the backup taken in
    :meth:`topology_setup`. This is done after each test, so each test starts
    with clear topology environment.

    When all tests for this topology are run, :meth:`topology_teardown` is
    called and the hosts are restored to the original state which backup was
    taken in :meth:`MultihostBackupHost.pytest_setup` so the environment is
    fresh for the next topology.

    .. note::

        It is possible to decorate methods, usually the custom implementation of
        :meth:`topology_setup` with :meth:`restore_vanilla_on_error`. This makes
        sure that the hosts are reverted to the original state if any of the
        setup calls fail.

        .. code-block:: python

            @BackupTopologyController.restore_vanilla_on_error
            def topology_setup(self, *kwargs) -> None:
                raise Exception("Hosts are automatically restored now.")
    """

    def __init__(self) -> None:
        super().__init__()

        self.backup_data: dict[MultihostBackupHost, Any | None] = {}
        """
        Backup data. Dictionary with host as a key and backup as a value.
        """

    def restore(self, hosts: dict[MultihostBackupHost, Any | None]) -> None:
        """
        Restore given hosts to their given backup.

        :param hosts: Dictionary (host, backup)
        :type hosts: dict[MultihostBackupHost, Any  |  None]
        :raises ExceptionGroup: If some hosts fail to restore.
        """
        errors = []
        for host, backup_data in hosts.items():
            if not isinstance(host, MultihostBackupHost):
                continue

            try:
                host.restore(backup_data)
            except Exception as e:
                errors.append(e)

        if errors:
            raise ExceptionGroup("Some hosts failed to restore to original state", errors)

    def restore_vanilla(self) -> None:
        """
        Restore to the original host state that is stored in the host object.

        This backup was taken when pytest started and we want to revert to this
        state when this topology is finished.
        """
        restore_data: dict[MultihostBackupHost, Any | None] = {}

        for host in self.hosts:
            if not isinstance(host, MultihostBackupHost):
                continue

            restore_data[host] = host.backup_data

        self.restore(restore_data)

    def topology_setup(self, *args, **kwargs) -> None:
        """
        Take backup of all topology hosts.
        """
        super().topology_setup(**kwargs)

        for host in self.hosts:
            if not isinstance(host, MultihostBackupHost):
                continue

            self.backup_data[host] = host.backup()

    def topology_teardown(self, *args, **kwargs) -> None:
        """
        Remove all topology backups from the hosts and restore the hosts to the
        original state before this topology.
        """
        try:
            for host, backup_data in self.backup_data.items():
                if not isinstance(host, MultihostBackupHost):
                    continue

                host.remove_backup(backup_data)
        except Exception:
            # This is not that important, we can just ignore
            pass

        self.restore_vanilla()

    def teardown(self, *args, **kwargs) -> None:
        """
        Restore the host to the state created by this topology in
        :meth:`topology_setup` after each test is finished.
        """
        self.restore(self.backup_data)

    @staticmethod
    def restore_vanilla_on_error(method):
        """
        Decorator. Restore all hosts to its original state if an exception
        occurs during method execution.

        :param method: Method to decorate.
        :type method: Any setup or teardown callback.
        :return: Decorated method.
        :rtype: Callback
        """

        @wraps(method)
        def wrapper(self: BackupTopologyController, *args, **kwargs):
            try:
                return self._invoke_with_args(partial(method, self))
            except Exception:
                self.restore_vanilla()
                raise

        return wrapper
